//
// Marina Drosou (mdrosou@cs.uoi.gr)
// University of Ioannina
//
// PrefSIENA Subscriber
//
// Last update: 12 Dec 2008
//

import java.util.Vector;
import java.util.Iterator;
import java.util.Collections;
import java.util.LinkedList;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.StringTokenizer;
import java.util.Comparator;
import java.util.Date;

import siena.*;

//
// ExtNotifications is a an extended version of notifications including their notification rank (aka score)
//
class ExtNotification {
	public Notification n;
	public double score;
	public long freshness;
	
	public ExtNotification(Notification notif, double sc, long fr)  {
		n = new Notification(notif);
		score = sc;
		freshness = fr;
	}
}

//
// Comparator used to compare ExtNotifications according to their score (used to rank received notifications for presentation reasons)
//
class NotificationComparator implements Comparator{
	public int compare(Object obj1, Object obj2) {
	ExtNotification node1 = (ExtNotification) obj1;
	ExtNotification node2 = (ExtNotification) obj2;
	
	if(node1.score < node2.score)
		return 1;
	else if(node1.score > node2.score)
		return -1;
	else
		return 0;
	}
}

//
// Subscriber class
//
public class Subscriber implements Notifiable {

	private static int num_of_servers = 2;								// number of servers to choose one to connect to
	private static ThinClient[] siena = new ThinClient[num_of_servers];	// the actual servers
	private static Statistics[] stats = new Statistics[num_of_servers]; // the corresponding statistics
	
	// variables used for the experiments' measurements
	private static LinkedList ReceivedNotifications = new LinkedList();
	private static int received_notifications = 0;	
	private static double sum_score =  0.0;
	private static double avg_score =  0.0;
	private static long sum_freshness = (long)0.0;
	private static long avg_freshness = (long)0.0;	
	private static int window = 0;	
	
	private static String newline = System.getProperty("line.separator");
	
	
	//
	// mdrosou - read subscriptions
	//
	public static Vector readData(String filename) {
		Vector filters = new Vector();
		
		try{
			// Read subscriptions from file and create the corresponding filters
			FileReader inFile = new FileReader(filename);
			BufferedReader bufRead = new BufferedReader(inFile);
			
	        String line;
			Filter f = new Filter();			
			
			boolean op_eq, op_lt, op_gt, op_ss;
			
			while((line = bufRead.readLine()) != null) {
				if(line.equals("")) {
					filters.add(f);
					f = new Filter();
					continue;
				}
				
				String tokens[] = {"","",""};
				int count = 0;
				op_eq = false; op_lt = false; op_gt = false; op_ss = false;
				
				StringTokenizer st = new StringTokenizer(line);
				while (st.hasMoreTokens())
					tokens[count++] = st.nextToken();	
				
				if(tokens[1].equals("=")) op_eq = true;
				if(tokens[1].equals("<")) op_lt = true;
				if(tokens[1].equals(">")) op_gt = true;
				if(tokens[1].equals("*")) op_ss = true;
				
				if(tokens[0].equals("score"))
					f.score = Double.parseDouble(tokens[1]);
				else if((op_eq || op_lt || op_gt || op_ss) == false)
					f.addConstraint(tokens[0], tokens[1]);
				else if(op_eq)
					f.addConstraint(tokens[0], Op.EQ, Double.parseDouble(tokens[2]));
				else if(op_gt)
					f.addConstraint(tokens[0], Op.GT, Double.parseDouble(tokens[2]));
				else if(op_lt)
					f.addConstraint(tokens[0], Op.LT, Double.parseDouble(tokens[2]));
				else
					f.addConstraint(tokens[0], Op.SS, tokens[2]);
		
	        }
	            
	    bufRead.close();
		} catch (Exception e){ 
			System.err.println("Error: " + e.getMessage());
		}
		return filters;
	}	
	
	//
	// mdrosou - stats notify
	//
	public void notify(Statistics s) {
        System.out.println("I just got this stat: server = " + s.server_index + " num_of_new_roots = " + s.num_of_new_roots + " num_of_poset_nodes = " + s.num_of_poset_nodes + " num_of_covered_nodes = " + s.num_of_covered_nodes);
		stats[s.server_index] = s;
    };
	
	//
	// mdrosou -  notify implementations
	//	
	public void notify(Notification e) {
		received_notifications++;
        System.out.println("I just got this event (# " + received_notifications + "):");
        System.out.println(e.toString());	
    };

    public void notify(Notification e, double s) {
		long fr;
		Date now = new Date();		
		received_notifications++;
		
		fr = now.getTime() - e.publication.getTime();
		sum_freshness += fr;
		sum_score += s;
		avg_score = (double)sum_score/received_notifications;
		avg_freshness = (long)sum_freshness/received_notifications;		
		ReceivedNotifications.add(new ExtNotification(e, s, fr));
        
		System.out.println("I just got this event (# " + received_notifications + ") with score " + s + ":" + newline + e.toString() + newline +
						   "Average score: " + avg_score + " Average freshness: " + avg_freshness + newline);
		System.out.flush();
    };
	
	public void notify(Notification [] s) { }
	
	//
	// mdrosou - chose server to connect to -t - private also
	//
	private static int chooseServer() {		
		int server = 0;
		double maxservres = -100000.0;
		double servres;
		
		for (int i=0;i<siena.length;i++) {
			servres = -1.0 * stats[i].num_of_new_roots + -1.0 * stats[i].num_of_poset_nodes + 1.0 * stats[i].num_of_covered_nodes;
			System.out.println(i + " " + servres);
			if(servres > maxservres) {
				maxservres = servres;
				server = i;
			}
		}
		return server;		
	}
	
	//
	// mdrosou - write results to file
	//
	synchronized private static void write_results(String suffix, boolean sorted) {
		ExtNotification notif;
		double sum = 0.0;
		long fr_min = 10000000, fr_max = -1;
		try{
			String f1 = "Measurements_" + suffix + ".txt";
			String f2 = "ReceivedNotifications_" + suffix + ".txt";
			
			FileWriter fstream = new FileWriter(f1);
			BufferedWriter out = new BufferedWriter(fstream);
			
			FileWriter fstream2 = new FileWriter(f2);
			BufferedWriter out2 = new BufferedWriter(fstream2);
			
			if(sorted) {
				NotificationComparator comp = new NotificationComparator();
				Collections.sort(ReceivedNotifications, comp);
			}
			
			Iterator li = ReceivedNotifications.iterator();
			while(li.hasNext()) {
				notif = (ExtNotification)li.next();
				sum+= notif.score;
				if(notif.freshness > fr_max)
					fr_max = notif.freshness;
				if(notif.freshness < fr_min)
					fr_min = notif.freshness;
				out2.write(notif.n.toString() + " " + notif.score + newline);
			}
			
			if(!sorted) {
				NotificationComparator comp = new NotificationComparator();
				Collections.sort(ReceivedNotifications, comp);
			}
			
			out.write(suffix + newline);
			out.write(newline + "Notifications received: " + received_notifications + newline);
			out.write("Average rank: " + (double)sum/received_notifications + newline);			
			out.write("Min rank: " + ((ExtNotification)ReceivedNotifications.getLast()).score + newline);
			out.write("Max rank: " + ((ExtNotification)ReceivedNotifications.getFirst()).score + newline);
			out.write("Average freshness: " + avg_freshness + newline);			
			out.write("Min freshness: " + fr_min + newline);
			out.write("Max freshness: " + fr_max + newline);
			
			LinkedList ReceivedNotificationsCopy = new LinkedList();
			ReceivedNotificationsCopy.addAll(ReceivedNotifications);
			
			Notification tmp, tmp2;
			ExtNotification etmp, etmp2;
			double distance = 0.0;
			double dist = 0.0;
			double mindist = 1000000.0;
			double maxdist = -1.0;
			
			li = ReceivedNotifications.iterator();
			Iterator li2;
			
			while(li.hasNext()) {
				etmp = (ExtNotification)li.next();
				tmp = (Notification)etmp.n;
				
				li2 = ReceivedNotificationsCopy.iterator();
				while(li2.hasNext()) {
					etmp2 = (ExtNotification)li2.next();
					tmp2 = (Notification)etmp2.n;
					
					if(ReceivedNotificationsCopy.indexOf(etmp2) > ReceivedNotifications.indexOf(etmp)){
						dist = tmp.notificationDistance(tmp2);					
						distance += dist;
						
						if(dist > maxdist)
							maxdist = dist;
						if(dist < mindist)
							mindist = dist;
						//out.write(tmp.toString() + newline + tmp2.toString() +newline + tmp.notificationDistance(tmp2) + newline + newline);
					}
				}
			}
			
			int received_notifications = ReceivedNotifications.size();
			double temp = (received_notifications * (received_notifications - 1))/(double)2.0;
			
			//out.write("Distance: " + distance + newline);		
			out.write("Normalized distance: " + (double)distance/temp + newline);	
			out.write("Min distance: " + mindist + newline);
			out.write("Max distance: " + maxdist + newline);
			
			out.close();
			out2.close();
		}catch (Exception e){ 
			System.err.println("Error: " + e.getMessage());
		}
	}

    public static void main(String[] args) {
		if(args.length != 4) {
			System.out.println("Usage: ?>Subscriber [subscriptions_file] [suffix] [wait time] [sorted]");
			System.exit(1);
		}
		
		window = 3;
		
		System.out.println("Subscriber");
		
		try {
			// start client
			Subscriber party = new Subscriber();	
		
			// define a list of servers
			siena[0] = new ThinClient("ka:localhost:2345");
			siena[1] = new ThinClient("ka:localhost:2347");			
		
			Vector filters = new Vector();			

			filters = readData(args[0]);	
		
			// collect all subscriptions to an array to later subscribe
			Filter [] ff = new Filter[filters.size()];
			int j = 0;
			Iterator si;
			for(si = filters.iterator(); si.hasNext();)
				ff[j++] = new Filter((Filter)si.next());		
		
			// collect all subscriptions to a pattern to send them to the servers
			// Pattern p = new Pattern(ff);
			
			// get statistics from all the servers
			// for (int i=0;i<siena.length;i++) {
				// stats[i] = new Statistics(10000, 10000, 10000, 10000);
				// siena[i].getStats(party, i, p);
			// }
		
			// wait for the servers to respond
			//Thread.sleep(4000);
		
			// chose server to connect to
			int server = 0; //chooseServer();
			
			//System.out.println("I chose this server: " + server);
				
			// disconnect from the other servers
			// for (int i=0;i<siena.length;i++)
				// if(i != server)
					// siena[i].shutdown();
		    
			// Thread.sleep(2000);
		
			// send subscriptions to the chosen server
			try {		
				for (int i=0; i<filters.size(); i++) {
					Thread.sleep(100);
					System.out.println("subscribing for " + ff[i].toString());
					siena[server].subscribe(ff[i], party);
				}
						
				try {
				    Thread.sleep(5000);
				} catch (java.lang.InterruptedException ex) {
				    System.out.println("interrupted"); 
				}
				
				// wait for notifications to arrive
				Thread.sleep(Integer.parseInt(args[2]));
				
				System.out.println("Asking server to write results to file.");
				String s = "IndermediateResults_" + args[1] + ".txt";
				siena[server].writeToFile(s);
				
				//System.out.println("unsubscribing");
				for (int i=0; i<filters.size(); i++) {
					Thread.sleep(50);
					//siena[server].unsubscribe(ff[i], party);
				}
		    } catch (SienaException ex) {
				System.err.println("Siena error:" + ex.toString());
		    }
	    
			Thread.sleep(1000);			
			
			System.out.println("writing results to file.");
			
			boolean sorted = false;
			
			if(args[3].equals("sorted"))		
				sorted = true;
			else
				sorted = false;			
			
			
			write_results(args[1], sorted);				
			
			System.out.println("shutting down.");			
			
			siena[server].shutdown();	
			Thread.sleep(3000);			
			System.exit(0);
		} catch (Exception ex) {
		    ex.printStackTrace();
		    System.exit(1);
		} 
    }
}
