//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.cs.colorado.edu/serl/siena/
//
//  Author: Antonio Carzaniga <carzanig@cs.colorado.edu>
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 1998-2002 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
//
// $Id: HierarchicalDispatcher.java,v 1.70 2006/05/09 17:26:47 carzanig Exp $
//

//
// mdrosou - PrefSIENA is an extension of siena for diverese and preference-aware notification delivery
// Implemented by Marina Drosou (mdrosou@cs.uoi.gr)
// All modified parts of the code have comments starting with the prefix "mdrosou" (like this one)
// Last update: 24 February 2009
//

package siena;

import siena.comm.*;

import java.util.Collection;
import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.HashMap;
import java.util.List;
import java.util.LinkedList;
import java.util.Iterator;
import java.util.ListIterator;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.BufferedReader;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;

// mdrosou - importing needed libraries
import java.util.Vector;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.io.FileWriter;
import java.io.BufferedWriter;

//
// mdrosou - in this version of Hierarchical Dispatcher the following changes are implemented:
// 	- subscriptions in the poset are associated with pairs of the form (subscriber, score) instead of just subscribers (see ExtSubscriber class).
//	- matching notifications are associated with the maximum score of the corresponding subscriptions (see addNotification function).
//	- there are 3 delivering policies for notification delivery. One can be choosed by altering the value of the <mode> variable of the HierarchicalDispatcher class (see publish functions).
//		- Periodic: Every time a notification is published it is buffered. Every <period_counter> publications, the top-k events for all subscribers (not just matching) are delivered.
//		- Sliding window: Every time a notification is published it is buffered. The windows of all subscribers are slided and their top-k events are delivered (if not already forwarded in the past).
//		- History: Every time a notification is published it is either delivered or not to the matching subscribers.
//

//
// mdrosou - <CandidateNotif> describes the form of the elements in the various lists used in the 
// periodic and slidibg window delivering policies. Each element consists of a notification,  its regular and diverse
// score and a flag to know if it has already been sent to the subscriber. Its expiration time is included in <notif>.
//
class CandidateNotif {
	public Notification notif;
	public double score;
	public double diverseScore;	
	public double diverseScore2;	
	public boolean already_sent;
	
	public CandidateNotif(double num, Notification e) {
	score = num;
	diverseScore = -1.0;
	diverseScore2 = -1.0;
	notif = new Notification(e);
	already_sent = false;
	}
	
	public CandidateNotif(CandidateNotif cn) {
	notif = new Notification(cn.notif);
	score = cn.score;
	diverseScore = cn.diverseScore;
	diverseScore2 = cn.diverseScore2;
	already_sent = cn.already_sent;
	}
	
	public String toString() {
	DateFormat formatter = new SimpleDateFormat("yyyy.MM.dd HH:mm:ss z");
	return notif.toString() + " [Score: " + Double.toString(score) + ", Diverse Score: " + Double.toString(diverseScore) +  "]";
	}
}

// mdrosou - comparators for the elements of the <CandidateNotif> lists
class NotifComparator implements Comparator{
	public int compare(Object obj1, Object obj2) {
	CandidateNotif node1 = (CandidateNotif) obj1;
	CandidateNotif node2 = (CandidateNotif) obj2;
	
	if(node1.score < node2.score)
		return 1;
	else if(node1.score > node2.score)
		return -1;
	else
		return 0;
	}
}

class DivNotifComparator implements Comparator{
	public int compare(Object obj1, Object obj2) {
	CandidateNotif node1 = (CandidateNotif) obj1;
	CandidateNotif node2 = (CandidateNotif) obj2;
	
	if(node1.diverseScore < node2.diverseScore)
		return 1;
	else if(node1.diverseScore > node2.diverseScore)
		return -1;
	else
		return 0;
	}
}

class DivNotifComparator2 implements Comparator{
	public int compare(Object obj1, Object obj2) {
	CandidateNotif node1 = (CandidateNotif) obj1;
	CandidateNotif node2 = (CandidateNotif) obj2;
	
	if(node1.diverseScore2 < node2.diverseScore2)
		return 1;
	else if(node1.diverseScore2 > node2.diverseScore2)
		return -1;
	else
		return 0;
	}
}

//
// mdrosou - <DivTable> is a data structure for the implemementation of the diversity heuristic of the history delivery policy
// The distances of the current top-k results of the user are maintained (those results may not be in the current window anymore
// but are the one seen by the user at the present time)
//
class DivTable {
	int k;
	double sigma;
	double[][] distances;
	public LinkedList<CandidateNotif> topevents = new LinkedList<CandidateNotif>();
	double diversity = 0.0;
	double ranksum = 0.0;
	public CandidateNotif removed = null;
	
	//
	// the first time k matching events are published, a DIVTAble is created. It stores the distances among these events.
	//
	public DivTable(int kvalue, double sigmavalue, LinkedList<CandidateNotif> events) {
		k = kvalue;
		sigma = sigmavalue;
		distances = new double[k][k+1];		// last column is the sum of the other columns
		topevents.addAll(events);
		
		// compute distances
		for(int i=0; i<k; i++) {
			distances[i][i] = 0.0;			
			for(int j=0; j<i; j++)
				distances[i][j] = distances[j][i] = events.get(i).notif.notificationDistance(events.get(j).notif);					
		}

		// compute sum of each row
		for(int i=0; i<k; i++) {
			double sum = 0.0;			
			for(int j=0; j<k; j++)
				sum += distances[i][j];
			distances[i][k] = sum;
		}		
		
		// compute average rank and diversity of topevents (there are always k events, so there is no need to normalize those quantities)
		for(int i=0; i<k; i++) {
			diversity += distances[i][k];
			ranksum += topevents.get(i).score;
		}		
		
		diversity /= 2.0;
	}
	
	//
	// check if a new notification increases the rank+diversity of the topevents. If so,  replace a top-event with this new notification 
	//	
	public boolean addNew(CandidateNotif cn) {
		boolean added = false;	
		
		// compute distance array for new event
		double[] newdist = new double[k+1];
		double sum = 0.0;
		for(int i = 0; i < k; i++) {
			newdist[i] = topevents.get(i).notif.notificationDistance(cn.notif);
			sum += newdist[i];				
		}
		newdist[k] = sum;
		
		// swap new notification with existing topevents and see if it improves the diversity
		double maximprovement = sigma * ranksum + (1 - sigma) * diversity;
		int position = -1;		
		
		for(int i = 0; i < k; i++) {
			double tmpimprovement = diversity - distances[i][k] + newdist[k] - topevents.get(i).notif.notificationDistance(cn.notif);
			if(tmpimprovement > maximprovement) {
				maximprovement = tmpimprovement;
				position = i;
			}		
		}
		
		removed = null;
		
		// if it improves the diversity, then replace the old topevent with the new one
		if(position != -1) {
			removed = topevents.get(position);
			
			for(int i = 0; i < k; i++) {
				distances[i][position] = distances[position][i] = newdist[i];
				distances[position][position] = 0.0;
				distances[i][k] -= topevents.get(i).notif.notificationDistance(topevents.get(position).notif);
				distances[i][k] += topevents.get(i).notif.notificationDistance(cn.notif);
			}
			
			distances[position][k] = newdist[k] - cn.notif.notificationDistance(topevents.get(position).notif);				
			
			diversity = 0.0;
			for(int i=0; i<k; i++)
				diversity += distances[i][k];				
			diversity /= 2.0;
				
			ranksum -= topevents.get(position).score;
			ranksum += cn.score;
			
			topevents.set(position, cn);
			
			added = true;	
		}		
		
		return added;	
	}
	
	
	//
	// replace an old top-event with a new one
	//	
	public void replace(CandidateNotif exp, CandidateNotif cn) {
		// compute distance array for new event
		double[] newdist = new double[k+1];
		double sum = 0.0;
		for(int i = 0; i < k; i++) {
			newdist[i] = topevents.get(i).notif.notificationDistance(cn.notif);
			sum += newdist[i];				
		}
		newdist[k] = sum;
		
		int position = topevents.indexOf(exp);
		
		removed = null;
		
		// if it improves the diversity, then replace the old topevent with the new one
		if(position != -1) {
			removed = topevents.get(position);
			for(int i = 0; i < k; i++) {
				distances[i][position] = distances[position][i] = newdist[i];
				distances[position][position] = 0.0;
				distances[i][k] -= topevents.get(i).notif.notificationDistance(topevents.get(position).notif);
				distances[i][k] += topevents.get(i).notif.notificationDistance(cn.notif);
			}
			
			distances[position][k] = newdist[k] - cn.notif.notificationDistance(topevents.get(position).notif);				
			
			diversity = 0.0;
			for(int i=0; i<k; i++)
				diversity += distances[i][k];				
			diversity /= 2.0;
				
			ranksum -= topevents.get(position).score;
			ranksum += cn.score;
			
			topevents.set(position, cn);
		}		
	}
	
	//
	// print <DivTable>
	//
	public String toString() {
		String result = topevents.toString();
		result = result.concat("\n\n");

		for (int i = 0; i < distances.length; i++) {
			for (int j = 0; j < distances[i].length; j++)
				result = result.concat("\t" + distances[i][j]);
			result = result.concat("\n");
		}
		
		result = result.concat("Diversity: " + diversity);
		return result;
	}	
}

//
// this is the abstraction of the subscriber used by the
// HierarchicalDispatcher.  It represents remote as well as local
// notifiable objects.  In addition to that, this object keeps track
// of failed attempts to contact the notifiable object so that
// HierarchicalDispatcher can periodically clean up its subscriber
// tables
//
abstract class Subscriber {
    public  short failed_attempts = 0;
    public  long first_error_time= 0;

    protected boolean suspended	= false;
    int refcount	= 0;
	
	// mdrosou
	public int k;		
	//public long total_div_time = (long)0.0;
	public int divcounter = 0;
	DivTable matrix = null;
	LinkedList results_per_window = new LinkedList();
	Set current_results = new HashSet();
	
	// mdrosou - <CandidateNotifications> is used to buffer notifications in the periodic and sliding window timing policies
	LinkedList CandidateNotifications = new LinkedList();	

	
	// mdrosou - get line seperator to use in output messages
	String newline = System.getProperty("line.separator");

    abstract public Object getKey();

    abstract public boolean isLocal();

    public byte[] getIdentity() { return null; }

    synchronized public void addRef() {
	++refcount;
    }

    synchronized public void removeRef() {
	if (refcount > 0) --refcount;
    }

    synchronized public boolean hasNoRefs() {
	return refcount == 0;
    }

    synchronized public void mapHandler(Interface i) { }

    synchronized public void suspend() {
	suspended = true;
    }

    synchronized public void resume() {
	suspended = false;
    }

    //
    // returns false whenever the notification mechanism failed, and
    // therefore this notifiable could be considered unreachable (am I
    // making sense?  this sentence might sound a bit "italian", but I
    // don't have time to fix it now)
    //
    abstract public boolean notify(SENPPacket pkt);

    abstract public void notify(Notification n);

    abstract public void notify(Notification [] s);

    protected void handle_notify_error(Exception ex) {
	if (failed_attempts == 0) 
	    first_error_time = System.currentTimeMillis();
	++failed_attempts;
	Logging.prlnerr("error (" + failed_attempts + ") notifying " 
			+ toString() + ": " + ex.toString());
    }

    synchronized public long getMillisSinceGood() {
	if (suspended || failed_attempts == 0) return 0;
	return System.currentTimeMillis() - first_error_time;
    }

    synchronized public short getFailedAttempts() {
	if (suspended) return 0;
	return failed_attempts;
    }

    public void shutdown() { }

    public String toString() {
	return getKey().toString();
    }

    public void writeIdentityTo(SENPWriter out) throws IOException {
	//
	// defined by subclasses as needed
	//
    }

    public void writeContactTo(SENPWriter out) throws IOException {
	//
	// defined by subclasses as needed
	//
    }
	
	//
	//  mdrosou - add a received notification  <e> to the subscriber's buffer with notification score equal to the max of the scores
	// of the subscriptions that cover <e>	
	//
	synchronized public boolean addNotification(double score, Notification e) {	
	
	// create new candidate notification
	CandidateNotif cn = new CandidateNotif(score, e);
	
	// if <e> is already in the buffer, keep the maximum score between <score> and the existing one
	// add new event to the buffer if it is not already there
	ListIterator li = CandidateNotifications.listIterator();
	boolean flag = true;
	
	while (li.hasNext()) {
		CandidateNotif tmp = (CandidateNotif) li.next();  
		
		//Comparing the sets themselves doesn't work for some reason...
		String s1 = tmp.notif.toString();
		String s2 = e.toString();		
		
		// if <e> already exists then don't add it again
		if(s1.equals(s2) ) {
			flag = false;
			
			// if the older score was lower, replace it
			if(score >= tmp.score)
				li.set(cn);
		}
	}
		// else add it to the buffer  (in the first position so newer elemets have higher priority when sorting later)
	if(flag)
		CandidateNotifications.addFirst(cn);
	// Logging.prlnlog(newline + newline + " current list: ");
	// li = CandidateNotifications.listIterator();
	// while (li.hasNext()) {
		// CandidateNotif tmp = (CandidateNotif) li.next(); 
		// Logging.prlnlog(tmp.notif.toString() + " " + tmp.score);
	// }
	
	return true;
	}	
	
	//
	// mdrosou - sort candidates in continuous mode and return the top-k ones
	// k == 0 not handled (in that case, I should forward everything)
	//
	synchronized public CandidateNotif[] sortCandidates_continuous(boolean diversify, double sigma) {	
	NotifComparator comp = new NotifComparator();
	DivNotifComparator divcomp = new DivNotifComparator();
	DivNotifComparator2 divcomp2 = new DivNotifComparator2();
	CandidateNotif[] s = new CandidateNotif[k];
	
	// get newest event
	CandidateNotif first = (CandidateNotif) CandidateNotifications.getFirst();	
	
	if(first.notif.toString().indexOf("Dummy") != -1)
		return s;	
	
	if(diversify) {
		//long startTimer = System.nanoTime();
		
		// find the divrank of the newest event
		double mindist = 100000.0;
		Iterator li = CandidateNotifications.iterator();
		
		if(li.hasNext())
			li.next();// next() not to get itself
		
		while(li.hasNext()) {
			CandidateNotif cn = (CandidateNotif) li.next();
			if(cn.already_sent == true && cn.score != -1.0) {		// -1.0 = dummy even
				
				double dist = cn.notif.notificationDistance(first.notif);
				
				//Logging.prlnerr(newline + newline + first.notif.toString() + " and " + cn.notif.toString() + "= " + dist);
				
				if(dist < mindist) {
					mindist = dist;			
				}
			}
		}
		
		if(mindist == 100000.0) //first event ever
			mindist = 0.0;
				
		
		first.diverseScore2 = mindist;						// mindist of each event at the time of its publication		
		
		//long endTimer = System.nanoTime();
		//total_div_time += endTimer - startTimer;
		
		LinkedList scoreList = new LinkedList();
		scoreList.addAll(CandidateNotifications);
		Collections.sort(scoreList, comp);					// sort all according to score
		
		LinkedList divList = new LinkedList();
		divList.addAll(CandidateNotifications);
		Collections.sort(divList, divcomp2);				// sort all according to mindist at the time of publication
		
		//startTimer = System.nanoTime();
		
		li = CandidateNotifications.iterator();
		while(li.hasNext()) {
			CandidateNotif tmp = (CandidateNotif) li.next();
			tmp.diverseScore = -(sigma * (double)scoreList.indexOf(tmp) + (1.0 - sigma) * (double)divList.indexOf(tmp));  // compute diverse score for all		
		}
		
		//endTimer = System.nanoTime();
		//total_div_time += endTimer - startTimer;
		//divcounter++;
	}
	
	// create a buffer copy and sort it
	LinkedList SortedCandidateNotifications = new LinkedList();
	SortedCandidateNotifications.addAll(CandidateNotifications);	
	
	if(diversify)		
		Collections.sort(SortedCandidateNotifications, divcomp);
	else
		Collections.sort(SortedCandidateNotifications, comp);
	
	// see if the top result has <first> in it	
	boolean send = false;
	
	//Logging.prlnerr(newline + newline + SortedCandidateNotifications.toString());	
	
	if(SortedCandidateNotifications.indexOf(first) < k && first.already_sent == false)
		send = true;
	
	if(send) {
		s[0] = first;
		first.already_sent = true;
		//Logging.prlnerr(newline + newline + "Sending: " + first.toString());	
	}
			
	return s;
	}			
	
	//
	// mdrosou - <diversify> takes an input list <list> and rearranges its elemets based on their score
	// and distance from each other to produce a diversified list
	//
	synchronized public LinkedList diversify_old(LinkedList list, double sigma) {
	LinkedList diverseList = new LinkedList();
	
	LinkedList scoreList = new LinkedList();
	LinkedList divList = new LinkedList();	
	
	CandidateNotif tmp, tmp2;
	Iterator li;
	
	scoreList.addAll(list);
	NotifComparator comp = new NotifComparator();
	Collections.sort(scoreList, comp);
	
	diverseList.add(list.getFirst());		// should be scoreList
	
	list.removeFirst();	
	int size = list.size();			// cannot use list.size(0 in the following for statement because of also using li.remove()
	
	// choose one element at a time and add it in <diverseList> until you have added k elements
	for(int i=0; i<size && i<k; i++) {		
		li = list.iterator();
		while(li.hasNext()) {
			tmp = (CandidateNotif) li.next();
			double mindist = 100000;
			
			Iterator dli = diverseList.iterator();
			while(dli.hasNext()) {
				tmp2 = (CandidateNotif) dli.next();
				double tmp_dis = tmp.notif.notificationDistance(tmp2.notif);
				if(tmp_dis < mindist)
					mindist = tmp_dis;	
			}
			
			tmp.diverseScore = mindist;		
		}
		
		// Logging.prlnerr(list.toString());
		// Logging.prlnerr(scoreList.toString());
		// Logging.prlnerr(divList.toString());
		
		divList.clear();
		divList.addAll(list);		
		DivNotifComparator divcomp = new DivNotifComparator();
		Collections.sort(divList, divcomp);		
		
		li = list.iterator();
		double minsum = 1000000;
		int index = -1;
		int maxIndex = -1;
		while(li.hasNext()) {
			index++;			
			tmp = (CandidateNotif) li.next();			
			double sum = sigma * (double)scoreList.indexOf(tmp) + (1.0 - sigma) * (double)divList.indexOf(tmp);    // sigma
			
			if(sum < minsum) {
				minsum = sum;
				maxIndex = index;
			}
		}
			
		diverseList.addLast(list.get(maxIndex));
		list.remove(maxIndex);
	}
	
	return diverseList;	
	}

	synchronized public LinkedList diversify(LinkedList list, double sigma) {
	LinkedList diverseList = new LinkedList();	// final result	
	LinkedList auxList = new LinkedList();		// auxilary list used at each step
	CandidateNotif tmp, tmp2;
	Iterator li;
	
	//String newline = System.getProperty("line.separator");
	//Logging.prlnerr("Initial:" + list.toString() + newline);
	
	// sort input list according to score
	NotifComparator comp = new NotifComparator();
	Collections.sort(list, comp);
	
	// add most prefered event to the result	
	diverseList.add(list.getFirst());
	list.removeFirst();
	
	// choose one element at a time and add it in <diverseList> until you have added k elements
	int size = list.size();			// cannot use list.size(0 in the following for statement because of also using li.remove()
	double distance[] = new double[size];
	
	for(int i=0; i<size; i++)
		distance[i] = 100000.0;		// max distance
		
	//long startTimer = System.nanoTime();
	
	for(int i=0; i<size && i<k; i++) {						
		li = list.iterator();
		int index = -1;
		
		while(li.hasNext()) {
			tmp = (CandidateNotif) li.next();
			index++;
			
			if(distance[index] != -1.0){			// removed elements
				double dist = tmp.notif.notificationDistance(((CandidateNotif)diverseList.getLast()).notif);
				if(dist < distance[index])
					distance[index] = dist;			
			}
			
			tmp.diverseScore = distance[index];		
		}
		
		//Logging.prlnerr("Step " + new Integer(i).toString() + newline + "Remaining:" + list.toString() + newline);
		
		auxList.clear();
		auxList.addAll(list);		
		DivNotifComparator divcomp = new DivNotifComparator();
		Collections.sort(auxList, divcomp);		
		
		//Logging.prlnerr("Aux:" + auxList.toString() + newline);
		
		li = list.iterator();
		double minsum = 1000000;
		int index2 = -1;
		int maxIndex = -1;
		while(li.hasNext()) {
			index2++;			
			tmp = (CandidateNotif) li.next();			
			double sum = sigma * (double)list.indexOf(tmp) + (1.0 - sigma) * (double)auxList.indexOf(tmp);    // sigma
			
			if(sum < minsum) {
				minsum = sum;
				maxIndex = index2;
			}
		}		
			
		diverseList.addLast(list.get(maxIndex));
		list.remove(maxIndex);
		distance[maxIndex] = -1.0;
		
		//Logging.prlnerr("Div:" + diverseList.toString() + newline + newline + newline);
	}
	
	//long endTimer = System.nanoTime();
	//total_div_time += endTimer - startTimer;
	//divcounter++;
	//Logging.prlnerr(newline + newline + "Diversity time: " + getIdentity().toString() + " " + new Long(total_time).toString() + " avg: " + new Long((long)(total_time/divcounter)).toString());
	
	return diverseList;	
	}		
	
	//
	// mdrosou - sort candidates in periodic mode and return the top-k ones
	// k == 0 not handled (in that case, I should forward everything)
	//
	synchronized public CandidateNotif[] sortCandidates_periodic(boolean diversify, double sigma) {
	NotifComparator comp = new NotifComparator();
	CandidateNotif[] s = new CandidateNotif[k];
	
	// sort buffer
	Collections.sort(CandidateNotifications, comp);
	
	// diversify buffer	
	if(diversify) {
		//long startTimer = System.nanoTime();	
		LinkedList DiverseCandidateNotifications = diversify(CandidateNotifications, sigma);	
		//long endTimer = System.nanoTime();
		//total_time += endTimer - startTimer;
		//divcounter++;
		//Logging.prlnerr(newline + newline + "Diversity time: " + getIdentity().toString() + " " + new Long(total_time).toString() + " avg: " + new Long((long)(total_time/divcounter)).toString());	
		CandidateNotifications.clear();
		CandidateNotifications.addAll(DiverseCandidateNotifications);
	}
	
	//Logging.prlnerr(CandidateNotifications.toString());
	
	// return the top-k notifications
	int counter = 0;
	int i = 0;
	
	Iterator li = CandidateNotifications.iterator();
	while(li.hasNext() && i<k) {
		CandidateNotif cn = (CandidateNotif) li.next();
		i++;
		s[counter] = cn;	
		counter++;	
	}
	
	// clear buffer, since all current events should be ignored in later periods
	CandidateNotifications.clear();
	
	return s;
	}	
	
	//
	// mdrosou - sort candidates in sliding window mode and return the top-k ones
	// k == 0 not handled (in that case, I should forward everything)
	//
	synchronized public CandidateNotif[] sortCandidates_slidingWindow(boolean diversify, double sigma) {
	NotifComparator comp = new NotifComparator();
	CandidateNotif[] s = new CandidateNotif[k];	
	
	// create a buffer copy and sort it
	LinkedList SortedCandidateNotifications = new LinkedList();
	SortedCandidateNotifications.addAll(CandidateNotifications);	
	Collections.sort(SortedCandidateNotifications, comp);
	
	//Logging.prlnerr(newline + newline + SortedCandidateNotifications.toString());
	
	// diversify buffer
	if(diversify) {
		//long startTimer = System.nanoTime();	
		LinkedList DiverseCandidateNotifications = diversify(SortedCandidateNotifications, sigma);	
		//long endTimer = System.nanoTime();
		//total_time += endTimer - startTimer;
		//divcounter++;
		//Logging.prlnerr(newline + newline + "Diversity time: " + getIdentity().toString() + " " + new Long(total_time).toString() + " avg: " + new Long((long)(total_time/divcounter)).toString());	
		SortedCandidateNotifications.clear();
		SortedCandidateNotifications.addAll(DiverseCandidateNotifications);
	}
	
	// locate the top-k notifications and only return the ones that have not been returned in the past
	int counter = 0;
	int i = 0;
	
	current_results.clear();
	Iterator li = SortedCandidateNotifications.iterator();
	while(li.hasNext() && i<k) {
		CandidateNotif cn = (CandidateNotif) li.next();
		i++;
		current_results.add(cn);
		if(!cn.already_sent) {		
			cn.already_sent = true;
			s[counter] = cn;		
			//Logging.prlnerr(newline + newline + "Sending: " + cn.toString());
			counter++;	
		}
	}
	
	results_per_window.add(new HashSet(current_results));
	return s;
	}	

	//
	// mdrosou - sort candidates in sliding-window2 mode and return the top-k ones (heuristic from continuous)
	// k == 0 not handled (in that case, I should forward everything)
	//
	synchronized public CandidateNotif[] sortCandidates_slidingWindow2(boolean diversify, double sigma) {	
	NotifComparator comp = new NotifComparator();
	DivNotifComparator divcomp = new DivNotifComparator();
	DivNotifComparator2 divcomp2 = new DivNotifComparator2();
	CandidateNotif[] s = new CandidateNotif[k];	
	
	if(diversify) {
		// get newest event 
		CandidateNotif first = (CandidateNotif) CandidateNotifications.getFirst();	
		
		// find the divrank of the newest event
		double mindist = 100000.0;
		Iterator li = CandidateNotifications.iterator();
		
		if(li.hasNext())
			li.next();// next() not to get itself
		
		while(li.hasNext()) {
			CandidateNotif cn = (CandidateNotif) li.next();
			if(cn.already_sent == true && cn.score != -1.0) {		// -1.0 = dummy event
				
				double dist = cn.notif.notificationDistance(first.notif);
				
				//Logging.prlnerr(newline + newline + first.notif.toString() + " and " + cn.notif.toString() + "= " + dist);
				
				if(dist < mindist) {
					mindist = dist;			
				}
			}
		}
		
		if(mindist == 100000.0) //first event ever
			mindist = 0.0;
				
		
		first.diverseScore2 = mindist;						// mindist of each event at the time of its publication		
		
		LinkedList scoreList = new LinkedList();
		scoreList.addAll(CandidateNotifications);
		Collections.sort(scoreList, comp);					// sort all according to score
		
		LinkedList divList = new LinkedList();
		divList.addAll(CandidateNotifications);
		Collections.sort(divList, divcomp2);				// sort all according to mindist at the time of publication
		
		li = CandidateNotifications.iterator();
		while(li.hasNext()) {
			CandidateNotif tmp = (CandidateNotif) li.next();
			tmp.diverseScore = -(sigma * (double)scoreList.indexOf(tmp) + (1.0 - sigma) * (double)divList.indexOf(tmp));  // compute diverse score for all		
		}
	}
	
	// create a buffer copy and sort it
	LinkedList SortedCandidateNotifications = new LinkedList();
	SortedCandidateNotifications.addAll(CandidateNotifications);	
	
	if(diversify)		
		Collections.sort(SortedCandidateNotifications, divcomp);
	else
		Collections.sort(SortedCandidateNotifications, comp);
	
	// locate the top-k notifications and only return the ones that have not been returned in the past
	int counter = 0;
	int i = 0;
	
	Iterator li = SortedCandidateNotifications.iterator();
	while(li.hasNext() && i<k) {
		CandidateNotif cn = (CandidateNotif) li.next();
		i++;
		if(!cn.already_sent) {		
			cn.already_sent = true;
			s[counter] = cn;	
			//Logging.prlnerr(newline + newline + "Sending: " + cn.toString());
			counter++;	
		}
	}
	return s;
	}

	//
	// mdrosou - sort candidates in history mode and return the top-k ones
	// k == 0 not handled (in that case, I should forward everything)
	//
	synchronized public CandidateNotif[] sortCandidates_history(CandidateNotif expired, int w, boolean diversify, double sigma) {	
	NotifComparator comp = new NotifComparator();
	CandidateNotif[] s = new CandidateNotif[k];
	boolean send = false;	
	
	// get newest event
	CandidateNotif first = (CandidateNotif) CandidateNotifications.getFirst();		
	if(first.notif.toString().indexOf("Dummy") != -1)
		return s;
		
	// count non-dummy elements
	int count = 0;
	Iterator lii = CandidateNotifications.iterator();
	while(lii.hasNext()) {
		CandidateNotif cn = (CandidateNotif) lii.next();
		if(cn.notif.toString().indexOf("Dummy") == -1) {
			count++;
		}
	}

	if(expired == null && count <= k)
		send = true;
		
	if(matrix == null && diversify) {
		// count non-dummy elements
		int counter = 0;
		LinkedList events = new LinkedList();
		Iterator li = CandidateNotifications.iterator();
		while(li.hasNext()) {
			CandidateNotif cn = (CandidateNotif) li.next();
			if(cn.notif.toString().indexOf("Dummy") == -1) {
				counter++;
				events.add(cn);
			}
		}
		
		if(counter == k) {	// first time there are k matching elements
			matrix = new DivTable(k, sigma, events);
			//Logging.prlnerr(newline + matrix.toString());	
		}		
	}
		
	// if(expired != null)
		// Logging.prlnerr(newline + newline + expired.toString());		
		
	if(!send && !diversify)	{
		if(expired != null && expired.already_sent && expired.notif.toString().indexOf("Dummy") == -1) { // <expired> was in top-k
			send = true;
		}
		
		if(!send) {	
			// create a buffer copy and sort it
			LinkedList SortedCandidateNotifications = new LinkedList();
			SortedCandidateNotifications.addAll(CandidateNotifications);	
			Collections.sort(SortedCandidateNotifications, comp);		
			
			// see if <first> is better than the worst event in the current top-k
			ListIterator li = SortedCandidateNotifications.listIterator(SortedCandidateNotifications.size());
			while(li.hasPrevious()) {
				CandidateNotif cn = (CandidateNotif) li.previous();
				if(cn.already_sent && cn.notif.toString().indexOf("Dummy") == -1) {
					if(cn.score <= first.score) {
						current_results.remove(cn);
						send = true;
						cn.notif.putAttribute("Dummy", 0);
						// Logging.prlnerr(newline + newline + cn.notif.toString());
					}
					break;
				}
			}
		}		
	}
	else if(!send && diversify)	{		
		if(expired != null && expired.already_sent && expired.notif.toString().indexOf("Dummy") == -1 && matrix.topevents.contains(expired)) { // <expired> was in top-k
			send = true;
			//Logging.prlnerr(newline + newline + "Added1: " + first.toString());
			//long startTimer = System.nanoTime();
			
			matrix.replace(expired, first);
			
			//long endTimer = System.nanoTime();
			//total_div_time += endTimer - startTimer;
			
			if(matrix.removed != null) {
				Iterator it = current_results.iterator();
				while(it.hasNext()) {
					CandidateNotif tmp = (CandidateNotif)it.next();
					if(tmp.toString().equals(matrix.removed.toString())) {
						//Logging.prlnerr(newline + "Removing3: " + tmp.toString());
						it.remove();	
						break;
					}
				}	
			}
		}
		
		if(!send && matrix != null) {
			//long startTimer = System.nanoTime();
			
			if(matrix.addNew(first)) {
				
				//long endTimer = System.nanoTime();
				//total_div_time += endTimer - startTimer;
				//Logging.prlnerr(newline + newline + "Added2: " + first.toString());
				
				Iterator it = current_results.iterator();
				while(it.hasNext()) {
					CandidateNotif tmp = (CandidateNotif)it.next();
					if(tmp.toString().equals(matrix.removed.toString())) {
						//Logging.prlnerr(newline + "Removing: " + tmp.toString());
						it.remove();	
						break;
					}
				}				
				//current_results.remove(matrix.removed);
				send = true;
				//Logging.prlnerr(newline + matrix.toString());	
			}
		}
	}	
	
	if(send) {		// <first> belongs to the top-k results
		s[0] = first;
		s[0].already_sent = true;
		current_results.add(s[0]);
		//Logging.prlnerr(newline + newline + "Sending: " + s[0].toString());			
	}

	results_per_window.add(new HashSet(current_results));

	return s;
	}			
}

class LocalSubscriber extends Subscriber {
    private Notifiable localobj;
  
    public LocalSubscriber(Notifiable n, int kk) {
	localobj = n;
	k = kk;
    }

    public Object getKey() {
	return localobj; 
    }

    public boolean isLocal() {
	return true;
    }

    public int hashCode() {
	return localobj.hashCode();
    }

    synchronized public void mapHandler(Interface i) {
	//
	// this should probably be signaled as an error---I don't
	// think it makes sense to remap a local notifiable.
	//
    }

    synchronized public boolean notify(SENPPacket pkt) {
	if (suspended) return true;
	try {
	    localobj.notify(pkt.event);
	    failed_attempts = 0;
	    return true;
	} catch (Exception ex) {
	    handle_notify_error(ex);
	    return false;
	}
    }

    synchronized public void notify(Notification n) {
	if (suspended) return;
	try {
	    localobj.notify(n);
	    failed_attempts = 0;
	} catch (Exception ex) {
	    handle_notify_error(ex);
	}
    }

    synchronized public void notify(Notification [] s) {
	if (suspended) return;
	try {
	    //
	    // here I purposely do not duplicate the sequence for
	    // efficiency reasons.  Clients should never modify
	    // objects passed through notify().
	    //
	    localobj.notify(s);
	    failed_attempts = 0;
	} catch (Exception ex) {
	    handle_notify_error(ex);
	}
    }
}

class RemoteSubscriber extends Subscriber {
    private String	identity_s = null;
    public final byte[] identity;
    private Interface	interf;  // I couldn't figure out a better
				 // identifier that would not clash
				 // with Java keywords (interface, if,
				 // int)
    private SENPPacket spkt;

    public RemoteSubscriber(byte [] id, Interface i, int kk) {
	identity = (byte[])id.clone();
	interf = i;
	interf.add_ref();
	spkt = new SENPPacket();
	k = kk;
    }

    public byte[] getIdentity() { return identity; }

    public Object getKey() {
 	return (identity_s != null) ? identity_s 
	    : (identity_s = new String(identity));
    }

    synchronized public void mapHandler(Interface i) {
	if (interf != null) interf.remove_ref();
	interf = i;
	interf.add_ref();
	suspended = false;
	first_error_time = 0;
	failed_attempts = 0;
    }

    public boolean isLocal() {
	return false;
    }

    public int hashCode() {
	return identity.hashCode();
    }

    synchronized public boolean notify(SENPPacket pkt) {
	if (suspended) return true;
	try {
	    interf.send(pkt.buf, pkt.encode());
	    failed_attempts = 0;
	    return true;
	} catch (Exception ex) {
	    handle_notify_error(ex);
	    return false;
	}
    }

    synchronized public void notify(Notification n) {
	if (suspended) return;
	try {
	    spkt.init();
	    spkt.id = HierarchicalDispatcher.master_id;
	    spkt.method = SENP.PUB;
	    spkt.event = n;
	    spkt.to = identity;
	    interf.send(spkt.buf, spkt.encode());
	    failed_attempts = 0;
	} catch (Exception ex) {
	    handle_notify_error(ex);
	}
    }

    synchronized public void notify(Notification [] s) {
	if (suspended) return;
	try {
	    spkt.init();
	    spkt.id = HierarchicalDispatcher.master_id;
	    spkt.method = SENP.PUB;
	    spkt.events = s;
	    spkt.to = identity;
	    interf.send(spkt.buf, spkt.encode());
	    failed_attempts = 0;
	} catch (Exception ex) {
	    handle_notify_error(ex);
	}
    }

    public void writeIdentityTo(SENPWriter out) throws IOException {
	out.write(0x53 /* 'S' */);
	out.write(identity);
	out.write(0x0a /* '\n' */);
    }

    public void writeContactTo(SENPWriter out) throws IOException {
 	out.write(0x49 /* 'I' */);
 	out.write(identity);
	out.write(0x0a /* '\n' */);
	interf.writeTo(out);
    }

    public void shutdown() {
	if (interf != null) {
	    interf.remove_ref();
	    interf = null;
	}
    }

    protected void finalize() throws Throwable {
	shutdown();
    }
}

//
// mdrosou - adding a new class to use in vector elements (in subscriptions' subscriber sets)
// An <ExtSubscriber> holds the subscriber and the score that the subscriber has
// assosiated with a specific subscription
//
class ExtSubscriber {
	private Subscriber subscr;
	private double score;
	
	public ExtSubscriber(Subscriber s, double num) {
		subscr = s;
		score = num;
	}
	
	public Subscriber getSubscr() {
		return subscr;
	}
	
	public double getScore() {
		return score;
	}
	
	public void setScore(double num) {
		score = num;
	}
}

class SubscriberIterator {
    private Iterator si;
    
    SubscriberIterator(Iterator i) {
	si = i;
    }

    public boolean hasNext() {
	return si.hasNext();
    }

    public Subscriber next() {
	return (Subscriber)si.next();
    }
}

// mdrosou - adding an iterator for the <ExtSubscriber>  class
class ExtSubscriberIterator {
    private Iterator si;
    
    ExtSubscriberIterator(Iterator i) {
	si = i;
    }

    public boolean hasNext() {
	return si.hasNext();
    }

    public ExtSubscriber next() {
	return (ExtSubscriber)si.next();
    }
}

//
// mdrosou - various modifications have been made to  the set classes (see original code to compare).
// The modifications mainly have to do with the use of the <ExtSubscriber> class instead of the <Subscriber> one.
//
class SSet {
    protected Set subs;
    
    public SSet() {
	subs = new HashSet();
    }
    
    // mdrosou - add the subscribers with their scores as in <s> combined with <e> (because <e> needs to be buffered for the periodic and sliding window timing policies)
	public boolean addAll(RefSSet s, Notification e) {
	ExtSubscriberIterator si;
	for(si = s.iterator(); si.hasNext();) {
	    ExtSubscriber exts = si.next();
		if(exts.getSubscr().addNotification(exts.getScore(), e)) {
			if(!(subs.add(exts.getSubscr()))) 
				return false;
		}
	}
	return true;
    }

    public boolean add(Subscriber s) {
	return subs.add(s);
    }

    public boolean contains(Subscriber s) {
	return subs.contains(s);
    }

    public boolean remove(Subscriber s) {
	return subs.remove(s);
    }

    public boolean isEmpty() {
	return subs.isEmpty();
    }

    public SubscriberIterator iterator() {
	return new SubscriberIterator(subs.iterator());
    }
}

class RefSSet {
	protected Vector subs;
	
    public RefSSet() {
	subs = new Vector();
    }

    // mdrosou
	public boolean addAll(RefSSet s) {
	boolean result = false;
	ExtSubscriberIterator si;
	for(si = s.iterator(); si.hasNext();) {
	    ExtSubscriber sub = si.next();
	    if(!(subs.contains(sub)))		
			if (subs.add(sub)) {
				sub.getSubscr().addRef();
				result = true;
	    }
	}
	return result;
    }

	// mdrosou - if the subscriber had already subscribed with some other score for that subscription  the old score is replaced
	public boolean add(ExtSubscriber s) {	
	Iterator si;
	for(si = subs.iterator(); si.hasNext();) {
		ExtSubscriber exts = (ExtSubscriber)si.next();
		if(exts.getSubscr() == s.getSubscr()) {
			exts.setScore(s.getScore());
			return true;
		}
	}
	
	if (subs.add(s)) {
	    s.getSubscr().addRef();
	    return true;
	} else {
	    return false;
	}
    }
	
	// mdrosou - checks wheather a subscriber set contains a subscriber
	public boolean contains(ExtSubscriber s) {
	Iterator si;
	for(si = subs.iterator(); si.hasNext();) {
		ExtSubscriber exts = (ExtSubscriber)si.next();
		if(exts.getSubscr() == s.getSubscr() && exts.getScore() == s.getScore())
			return true;
	}
	return false;	
	}
	
	// mdrosou 
	public boolean contains(Subscriber s) {
	Iterator si;
	for(si = subs.iterator(); si.hasNext();) {
		ExtSubscriber exts = (ExtSubscriber)si.next();
		if(exts.getSubscr() == s)
			return true;
	}
	return false;
	}

    public boolean remove(Subscriber s) {
	Iterator si;
	for(si = subs.iterator(); si.hasNext();) {
		ExtSubscriber exts = (ExtSubscriber)si.next();
		if(exts.getSubscr() == s) {
			subs.remove(exts); 							//does this work?
			s.removeRef();
			return true;
		}			
	}
	return false;
    }
	
	public boolean isEmpty() {
	return subs.isEmpty();
    }

    // mdrosou 
	public ExtSubscriberIterator iterator() {
	return new ExtSubscriberIterator(subs.iterator());
    }

    protected void finalize() throws Throwable {
	ExtSubscriberIterator si;
	for(si = this.iterator(); si.hasNext();)
	    si.next().getSubscr().removeRef();
    }
}

class Subscription {
    public Set		preset;
    public Set		postset;

    public final Filter	filter;
    public RefSSet	subscribers;

    public Subscription(Filter f) {
	preset = new HashSet();
	postset = new HashSet();
	filter = new Filter(f);
	subscribers = new RefSSet(); 
    }

    public Subscription(Filter f, Subscriber s) {
	preset = new HashSet();
	postset = new HashSet();
	filter = new Filter(f);
	subscribers = new RefSSet();
	// mdrosou
	ExtSubscriber exts = new ExtSubscriber(s, f.score);
	subscribers.add(exts);
    }

    public String toString() {
	Iterator i;
	StringBuffer sb = new StringBuffer();
	sb.append("SUB: " + Integer.toString(hashCode()));
	sb.append("\npreset:");
	for(i = preset.iterator(); i.hasNext();)
	    sb.append(" "+Integer.toString(i.next().hashCode()));
	sb.append("\npostset:");
	for(i = postset.iterator(); i.hasNext();)
	    sb.append(" "+Integer.toString(i.next().hashCode()));
	sb.append("\n" + filter.toString());
	sb.append("\nsubscribers:");
	ExtSubscriberIterator si;
	for(si = subscribers.iterator(); i.hasNext();)
	    sb.append(" "+Integer.toString(i.next().hashCode()));
	sb.append("\n");
	return sb.toString();
    }

    public void writeTo(SENPWriter out) throws IOException {
	// mdrosou
	String newline = System.getProperty("line.separator");
	out.write(newline);
	
	out.write(0x46 /* 'F' */);
	out.encodeSimple(filter);
	//out.write(0x0a /* '\n' */);
	
/* 	ExtSubscriberIterator si;	
	
	// mdrosou - modifying printing format
	ExtSubscriber i;	
	for(si = subscribers.iterator(); si.hasNext();) {
		i = si.next();
	    i.getSubscr().writeIdentityTo(out);
		out.write(" (" + Double.toString(i.getScore()) + ") ");
	} */
	
	out.write(newline);
    }
	
}

class Poset {
    private int		mods_since_save = 0;
    private int		time_since_save = 0;
    
	 // mdrosou - roots was private
	public Set		roots;
	
	// mdrosou
	public int num_checked_nodes = 0;
    
    public Poset() {
	roots = new HashSet();
    }

    public void clear() {
	roots.clear();
    }

    public boolean is_root(Subscription s) {
	return s.preset.isEmpty();
    }

    private boolean empty() {
	return roots.isEmpty();
    }

    public Iterator rootsIterator() {
	return roots.iterator();
    }

    private void insert(Subscription new_sub, 
			Collection pre, Collection post) {
	//
	// inserts new_sub into the poset between pre and post.  the
	// connections are rearranged in order to maintain the
	// properties of the poset
	//
	Subscription x; // x always represents something in the preset
	Subscription y; // y has to do with the postset
	Iterator xi, yi;

	if (pre.isEmpty()) {
	    roots.add(new_sub);
	    roots.removeAll(post);
	} else {
	    xi = pre.iterator();
	    while(xi.hasNext()) {
		x = (Subscription)xi.next();
		yi = post.iterator();
		while(yi.hasNext()) 
		    disconnect(x, (Subscription)yi.next());
		connect(x, new_sub);
	    }
	}
	yi = post.iterator();
	while(yi.hasNext()) {
	    y = (Subscription)yi.next();
	    connect(new_sub, y);
	}
	++mods_since_save;
    }

    private void disconnect(Subscription x, Subscription y) {
	if (x.postset.remove(y))
	    y.preset.remove(x);
    }

    private void connect(Subscription x, Subscription y) {
	if (x.postset.add(y))
	    y.preset.add(x);
    }

    public Set remove(Subscription s) {
	//
	// removes s from the poset returning the set of root
	// subscription uncovered by s
	//
	Set result = new HashSet();
	Subscription x, y;
	Iterator xi, yi;
	//
	// 1. disconnect s from every successor of s but maintains s.postset
	//
	yi = s.postset.iterator();
	while(yi.hasNext()) {
	    y = (Subscription)yi.next();
	    y.preset.remove(s);
	}

	if (s.preset.isEmpty()) {
	    //
	    // 2.1 if s is a root subscription, adds as a root every
	    // successor that remains with an empty preset, i.e.,
	    // every subscription that was a successor of s and that
	    // is now a root subscription...
	    //
	    yi = s.postset.iterator();
	    while(yi.hasNext()) {
		y = (Subscription)yi.next();
		if (y.preset.isEmpty()) {
		    roots.add(y);
		    result.add(y);
		}
	    }
	    roots.remove(s);
	} else {
	    //
	    // 2.2 disconnects every predecessor of s thereby reconnecting
	    // predecessors to successors. A predecessor X is re-connected
	    // to a successor Y only if X does not have an immediate
	    // successor X' that covers Y (see is_indirect_successor).
	    //
	    xi = s.preset.iterator();
	    while(xi.hasNext()) {
		x = (Subscription)xi.next();
		x.postset.remove(s);
		yi = s.postset.iterator();
		while(yi.hasNext()) {
		    y = (Subscription)yi.next();
		    if (!is_indirect_successor(x, y))
			connect(x, y);
		}
	    }
	}
	return result;
    }

    private boolean is_indirect_successor(Subscription x, Subscription y) {
	//
	// says whether x indirectly covers y.
	//
	Iterator i = x.postset.iterator();
	while(i.hasNext()) 
	    if (Covering.covers(((Subscription)i.next()).filter, y.filter))
		return true;
	return false;
    }

    private Set predecessors(Filter f, Subscriber s) {
	//
	// computes the set of immediate predecessors of filter f that
	// do not contain subscriber s.  If the poset contains any
	// subscription covering f that already contains s, then this
	// function returns null.  Otherwise, it returns the
	// collection of predecessors of f.  If the poset contains a
	// filter f'=f, the result will be a set containing that
	// (only) subscription.
	// 
	LinkedList to_visit = new LinkedList();
	Set visited = new HashSet();
	Subscription sub, y;
	Iterator i = roots.iterator();
	boolean found_lower;

	while(i.hasNext()) {
	    sub = (Subscription)i.next();
	    if (Covering.covers(sub.filter, f)) {
		if (sub.subscribers.contains(s)) {
		    return null;
		} else {
		    to_visit.addLast(sub);
		}
	    }
	}
	Set result = new HashSet();

	ListIterator li;
	while((li = to_visit.listIterator()).hasNext()) {
	    sub = (Subscription)li.next();
	    li.remove();
	    i = sub.postset.iterator();
	    found_lower = false;
	    while(i.hasNext()) {
		y = (Subscription)i.next();
		if (visited.add(y)) {
		    if (Covering.covers(y.filter, f)) {
			found_lower = true;
			if (sub.subscribers.contains(s)) {
			    return null;
			} else {
			    to_visit.addLast(y);
			}
		    }
		} else if (!found_lower) {
		    if (Covering.covers(y.filter, f))
			found_lower = true;
		}
	    }
	    if(!found_lower) result.add(sub); 
	}
	return result;
    }
	
	// mdrosou - handling the subscriber as an extended one (different implementation of contains())
    private Set predecessors(Filter f, ExtSubscriber s) {
	//
	// computes the set of immediate predecessors of filter f that
	// do not contain subscriber s.  If the poset contains any
	// subscription covering f that already contains s, then this
	// function returns null.  Otherwise, it returns the
	// collection of predecessors of f.  If the poset contains a
	// filter f'=f, the result will be a set containing that
	// (only) subscription.
	// 
	LinkedList to_visit = new LinkedList();
	Set visited = new HashSet();
	Subscription sub, y;
	Iterator i = roots.iterator();
	boolean found_lower;

	while(i.hasNext()) {
	    sub = (Subscription)i.next();
	    if (Covering.covers(sub.filter, f)) {
		if (sub.subscribers.contains(s)) {
		    return null;
		} else {
		    to_visit.addLast(sub);
		}
	    }
	}
	Set result = new HashSet();

	ListIterator li;
	while((li = to_visit.listIterator()).hasNext()) {
	    sub = (Subscription)li.next();
	    li.remove();
	    i = sub.postset.iterator();
	    found_lower = false;
	    while(i.hasNext()) {
		y = (Subscription)i.next();
		if (visited.add(y)) {
		    if (Covering.covers(y.filter, f)) {
			found_lower = true;
			if (sub.subscribers.contains(s)) {
			    return null;
			} else {
			    to_visit.addLast(y);
			}
		    }
		} else if (!found_lower) {
		    if (Covering.covers(y.filter, f))
			found_lower = true;
		}
	    }
	    if(!found_lower) result.add(sub); 
	} 
	return result;
    }

    public SSet matchingSubscribers(Notification e) {
	//
	// computes the set of subscribers that are interested in e.
	// This includes the subscribers of all the subscriptions in
	// the poset that match e
	//
	SSet result = new SSet();
	Iterator i = roots.iterator();
	LinkedList to_visit = new LinkedList();
	Set visited = new HashSet();
	Subscription sub;
	
	// mdrosou
	num_checked_nodes += roots.size();

	while(i.hasNext()) {
	    sub = (Subscription)i.next();
	    if (Covering.apply(sub.filter, e)) {
		to_visit.addLast(sub);
		
		// mdrosou - if the subscriber is added, <e> is buffered for him
		result.addAll(sub.subscribers, e);
	    }
	}

	ListIterator li;
	while((li = to_visit.listIterator()).hasNext()) {
	    sub = (Subscription)li.next();
	    li.remove();
	    i = sub.postset.iterator();
	    while(i.hasNext()) {
		// mdrosou
		num_checked_nodes += 1;
		
		Subscription y = (Subscription)i.next();
		if (visited.add(y) && Covering.apply(y.filter, e)) {
		    to_visit.addLast(y);
		    result.addAll(y.subscribers, e);
		}
	    }
	}
	return result;
    }

    public Set successors(Filter f, Collection pred) {
	//
	// given a filter f and a set pred, the set of the immediate
	// predecessors of f in the poset, computes the set of
	// immediate successors of f in the poset.
	//
	// Idea: I must walk through the sub-poset of this poset
	// that is covered by the set of predecessors (the whole poset
	// if there are no predecessors), looking for filters f1, f2,
	// ..., fn that are covered by f.  I do that by using a queue
	// of filters to_visit.
	// 
	// to_visit is initialized with the element of pred or with
	// the root filters. For every f' in to_visit, if f' is covered
	// by f then I remove f' from to_visit and I add it to the
	// result.  If f' is not covered by f, then I add all its
	// successors that I haven't visited to to_visit.
	//
	LinkedList to_visit = new LinkedList();
	Set visited = new HashSet();
	Subscription sub, y;
	Iterator i;
	//
	// initialize to_visit
	//
	if (pred == null || pred.isEmpty()) {
	    to_visit.addAll(roots);
	} else {
	    i = pred.iterator();
	    while(i.hasNext()) {
		to_visit.addLast((Subscription)i.next());
	    }
	}
	visited.addAll(to_visit);
	Set result = new HashSet();
	ListIterator li;
	while((li = to_visit.listIterator()).hasNext()) {
	    sub = (Subscription)li.next();
	    li.remove();
	    if (Covering.covers(f, sub.filter)) {
		result.add(sub);
	    } else {
		i = sub.postset.iterator();
		while(i.hasNext()) {
		    y = (Subscription)i.next();
		    if (visited.add(y)) {
			to_visit.addLast(y);
		    }
		}
	    }
	}

	return result;
    }

    public Subscription insert_subscription(Filter f, Subscriber s) {
	//
	// inserts a subscription in the poset
	//
	//Set pred = predecessors(f,s);
	// mdrosou - using ExtSubscriber instead
	ExtSubscriber exts = new ExtSubscriber(s, f.score);
	Set pred = predecessors(f,exts);
	
	Subscription sub; 
	if (pred == null) return null;

	if (pred.size() == 1) {
	    sub = (Subscription)pred.iterator().next();
	    if (Covering.covers(f,sub.filter)) {
		//
		// pred contains exactly f, so simply add s to the set
		// of subscribers
		//
		// mdrosou
		sub.subscribers.add(exts);
		clear_subposet(sub, s);
		return sub;
	    }
	}
	sub = new Subscription(f, s);
	insert(sub, pred, successors(f, pred));
	clear_subposet(sub, s);
	return sub;
    }

    //
	// mdrosou - commented this function out because we do not want to remove covered subscriptions
	// since they may have higher scores.
	//
	public void clear_subposet(Subscription start, Subscriber s) {
	//
	// removes subscriber s from all the subscriptions covered by
	// start, excluding start itself.  This also removes
	// subscriptions that remain with no subscribers.
	//	
	// LinkedList to_visit = new LinkedList();
	// Set visited = new HashSet();

	// to_visit.addAll(start.postset);

	// ListIterator li;
	// while((li = to_visit.listIterator()).hasNext()) {
	    // Subscription sub = (Subscription)li.next();
	    // li.remove();
	    // if (visited.add(sub)) {
		// if (sub.subscribers.remove(s)) {
		    // if (sub.subscribers.isEmpty()) remove(sub);
		// } else {
		    // to_visit.addAll(sub.postset);
		// }
	    // }
	// }
    }

    public Set to_remove(Filter f, Subscriber s) {
	//
	// removes subscriber s from the subscriptions covered by f.
	// If f==null, it removes s from all the subscriptions in the
	// poset. Returns the set of empty subscriptions, i.e., those
	// that remain with no subscribers.
	//
	Set result = new HashSet();
	LinkedList to_visit = new LinkedList();
	Subscription sub;

	if (f == null) {
	    //
	    // f==null ==> universal filter (same thing as BYE)
	    // so my starting point is the set of root subscriptions
	    //
	    to_visit.addAll(roots);
	} else {
	    Set pred = predecessors(f, s);
	    if (pred != null && pred.size() == 1) {
		sub = (Subscription)pred.iterator().next();
		if (Covering.covers(f,sub.filter)) {
		    //
		    // pred contains exactly f, so remove s and see if
		    // f remains with no subscribers
		    //
		    if (sub.subscribers.remove(s)) {
			if (sub.subscribers.isEmpty()) 
			    result.add(sub);
		    }

		    return result;
		}
	    }
	    to_visit.addAll(successors(f, pred));
	}

	Set visited = new HashSet();
	ListIterator li;

	while((li = to_visit.listIterator()).hasNext()) {
	    sub = (Subscription)li.next();
	    li.remove();
	    if (visited.add(sub)) {
		if (sub.subscribers.remove(s)) {
		    if (sub.subscribers.isEmpty()) 
			result.add(sub);
		} else {
		    to_visit.addAll(sub.postset);
		}
	    }
	}
	return result;
    }

    public void writeTo(SENPWriter out) throws IOException {
	//
	// prints the poset
	//
	LinkedList to_visit = new LinkedList();
	Set visited = new HashSet();

	to_visit.addAll(roots);
	
	// mdrosou 
	String newline = System.getProperty("line.separator");
	out.write(newline + newline);
	
	ListIterator li;
	while((li = to_visit.listIterator()).hasNext()) {
	    Subscription sub = (Subscription)li.next();
	    li.remove();
	    if (visited.add(sub)) {
		sub.writeTo(out);
		to_visit.addAll(sub.postset);
	    }
	}
    }
	
	// mdrosou - find poset size
	public int countSize() {
	LinkedList to_visit = new LinkedList();
	Set visited = new HashSet();
	int counter = 0;

	to_visit.addAll(roots);
	
	ListIterator li;
	while((li = to_visit.listIterator()).hasNext()) {
	    Subscription sub = (Subscription)li.next();
	    li.remove();
	    if (visited.add(sub)) {
		counter++;
		to_visit.addAll(sub.postset);
	    }
	}
	return counter;
    }
	
	//
	// mdrosou - the follwing functions collect statistics about the poset which
	// will be used to cluster the subscribers. I don't think they should be synchronized
	// since the function that calls them already is
	//
	
	//
	// mdrosou - count how many of the filters stored in <p> will be added as new roots to the poset
	// (currently does not consider covering among the filters in <p>. I should do that by copying <roots>
	// to a tmp value and add filters to it as well.
	//	
	public int countNewRoots(Pattern p) {
	int counter = 0;
	boolean flag = false;
	Set new_roots = new HashSet();
	
	new_roots.addAll(roots);
	
	for(int i=0;i<p.filters.length;i++) {
		Iterator j = new_roots.iterator();
		while(j.hasNext()) {
			Subscription sub = (Subscription)j.next();
			flag = false;
			if (Covering.covers(sub.filter, p.filters[i])) {
				Logging.prlnlog(p.filters[i].toString() + " is covered by " + sub.filter.toString());
				flag = true;
				counter++;
			}
		}
		if(!flag) {
		new_roots.add(new Subscription(p.filters[i]));
		Logging.prlnlog(p.filters[i].toString() + " is added as root ");
		}
	}
	return p.filters.length - counter;
	}
	
	//
	// mdrosou - count the existing subsciptions of the poset which will be covered by the new ones
	// The more they are, the more efficient future pruning will probably be
	//	
	public int countCoveredSubscriptions(Pattern p) {	
	Set preds, succs;
	LinkedList to_visit = new LinkedList();
	Set result = new HashSet();
	Set tmpresult = new HashSet();  // mdrosou - not needed, just for printing
	Subscription sub;
	
	for(int i=0;i<p.filters.length;i++) {
			tmpresult.clear();
			
			preds = predecessors(p.filters[i], (Subscriber)null);
			succs = successors(p.filters[i], preds);
			to_visit.addAll(succs);
			result.addAll(succs);
			
			tmpresult.addAll(succs);			
			
			ListIterator li;
			while((li = to_visit.listIterator()).hasNext()) {
				sub = (Subscription)li.next();
				li.remove();
				to_visit.addAll(sub.postset);
				result.addAll(sub.postset);
				
				tmpresult.addAll(sub.postset);
			}
			
			Logging.prlnlog(p.filters[i].toString() + " covers ");
			Iterator j = tmpresult.iterator();
			while(j.hasNext()) {
				sub = (Subscription)j.next();
				Logging.prlnlog(sub.toString());
			}
			
	}
	
	Logging.prlnlog("Final: ");
	Iterator j = result.iterator();
	while(j.hasNext()) {
		sub = (Subscription)j.next();
		Logging.prlnlog(sub.toString());
	}
	return result.size();
	}
	
	//
	// mdrosou - number of subscriptions already in the poset
	// counted for load balancing
	//	
	public int countSubscriptions() {
	LinkedList to_visit = new LinkedList();
	Set visited = new HashSet();
	Subscription sub, y;
	Iterator i;
	
	to_visit.addAll(roots);
	visited.addAll(roots);
	
	ListIterator li;
	while((li = to_visit.listIterator()).hasNext()) {
		sub = (Subscription)li.next();
		li.remove();		
		i = sub.postset.iterator();
	    while(i.hasNext()) {
			y = (Subscription)i.next();
			if (visited.add(y))
				to_visit.addLast(y);
		}
	}
	
	Logging.prlnlog("Final: ");
	Iterator j = visited.iterator();
	while(j.hasNext()) {
		sub = (Subscription)j.next();
		Logging.prlnlog(sub.toString());
	}
	
	return visited.size();
	}
	
}

class EmptyPatternException extends SienaException {
    // ...work in progress...
}

//
// we implement pattern recognition with PatternMatcher.  A
// PatternMatcher is a ``parser'' of event sequences.  A
// PatternMatcher receives notifications (single events) from Siena
// and holds a list of subscribers.  When a PatternMatcher recognizes
// a pattern, it notifies the list of subscribers passing the matching
// sequence of events.
//
class PatternMatcher implements Notifiable {
    public RefSSet		subscribers	= null;
    public Pattern		pattern		= null;
    private LinkedList		notifications	= null;

    public PatternMatcher(Pattern p) {
	pattern = new Pattern(p);
	notifications = new LinkedList();
	subscribers = new RefSSet();
    }
	
    // mdrosou - adding new type of notification
	public void notify(Statistics s) {}
	public void notify(Notification n, double score) {}
	
	public void notify(Notification s[]) {}

    public void notify(Notification n) {
	//
	// This method receives the elementary components of a pattern
	// and does the matching.
	//
	// WARNING: this is my own naive matching algorithm.  It is
	// certainly not optimized and it might not even be correct!
	// I should look up Knuth-Morris-Pratt matching algorithm,
	// but I have no time now...
	//
	// ...work in progress... bigtime.
	//
	int curr = notifications.size();
	notifications.addLast(new Notification(n));
	//
	// pattern.length must be > 0 (see constructor)
	//
	if (Covering.apply(pattern.filters[curr], n)) {
	    if (++curr == pattern.filters.length) {
		//
		// MATCHED!  builds the array of notifications,
		//
		Notification sequence[] = new Notification[curr];
		while(--curr >= 0) 
		    sequence[curr] = (Notification)notifications.removeLast();
		//
		// notify subscribers
		//
		// mdrosou
		ExtSubscriberIterator i = subscribers.iterator();
		while(i.hasNext()) {
		    Subscriber s = i.next().getSubscr();
		    s.notify(sequence);
		}
	    }
	} else {
	    //
	    // no match here, I've got to backtrack...  try to cut the
	    // head of the queue of notifications and match a
	    // shorter (most recent) prefix of filters.
	    //
	    for(notifications.removeFirst(); !notifications.isEmpty(); 
		notifications.removeFirst()) {
		ListIterator li = notifications.listIterator();
		for(curr = 0; li.hasNext(); ++curr)
		    if (!Covering.apply(pattern.filters[curr], 
					(Notification)li.next())) break;
		if (!li.hasNext()) {
		    //
		    // found a shorter prefix
		    //
		    return;
		}
	    }
	}
    }
}

class Interface {
    private final byte []	address;
    private final PacketSender	sender;
    private int			refs;

    Interface(byte [] h, PacketSender s) {
	address = (byte[])h.clone();
	sender = s;
	refs = 0;
    }

    synchronized public void add_ref() {
	++refs;
    }

    synchronized public void remove_ref() {
	if (refs > 0) --refs;
    }

    synchronized public boolean has_no_refs() {
	return refs == 0;
    }

    public void writeTo(SENPWriter out) throws IOException {
	out.write(0x48 /* 'H' */);
	out.write(address);
	out.write(0x0a /* '\n' */);
    }

    public void send(byte[] packet) throws PacketSenderException {
	sender.send(packet);
    }

    public void send(byte[] packet, int len) throws PacketSenderException {
	sender.send(packet, len);
    }

    public void send(byte[] packet, int offset, int len) 
	throws PacketSenderException {
	sender.send(packet, offset, len);
    }

    public void shutdown() throws PacketSenderException {
	sender.shutdown();
    }
}

class BadStorageFormat extends SienaException {
    BadStorageFormat(String s) {
	super(s);
    }
}

//
// this class is not intended for general use outside
// HierarchicalDispatcher.  In fact, I could have implemented
// everything directly within HierarchicalDispatcher.  In particular,
// it is HierarchicalDispatcher that performs the necessary
// concurrency control for ContactsTable.
//
class ContactsTable {
    // mdrosou -was private
	public HashMap contacts = new HashMap();	
    //
    // Note on Concurrency Control:
    //
    // in general, all these methods should be synchronized, but in
    // this case I prefer to control concurrency directly at the
    // HierarchicalDispatcher level.
    //
    public void clear() {
	contacts.clear();
    }

    public void writeTo(SENPWriter out) throws IOException {
	for(Iterator i = contacts.entrySet().iterator(); i.hasNext();) 
	    ((Subscriber)((Map.Entry)i.next()).getValue()).writeContactTo(out);
    }

    public void remove(Subscriber s) {
	if (contacts.remove(s.getKey()) != null)
	    Logging.prlnlog("removing " + s + " from contacts list");
    }

    public void put(Subscriber s) {
	
	contacts.put(s.getKey(), s);
    }

    public void put(Notifiable n) {
	contacts.put(n, n);	
    }

    public Subscriber get(String id) {
	return (Subscriber)contacts.get(id);
    }

    public Subscriber get(Notifiable n) {
	return (Subscriber)contacts.get(n);
    }

    public void cleanup() {
	Map.Entry e;
	for(Iterator i = contacts.entrySet().iterator(); i.hasNext();) {
	    e = (Map.Entry)i.next();
	    Subscriber sub = (Subscriber)e.getValue();
	    if (sub.hasNoRefs()) {
		sub.shutdown();
		Logging.prlnlog("contacts cleanup: removing subscriber " 
				+ e.getKey());
		i.remove(); 
		// assuming that this actually removes the
		// mapping from the original map---not
		// from the set returned by entrySet()
	    }
	}
    }
	
}

//
// this class is not intended for general use outside
// HierarchicalDispatcher.  In fact, I could have implemented
// everything directly within HierarchicalDispatcher.  In particular,
// it is HierarchicalDispatcher that performs the necessary
// concurrency control for IFManager.
//
class IFManager {
    static private PacketSenderFactory	default_factory 
					  = new GenericSenderFactory();

    private PacketSenderFactory		sender_factory = default_factory;
    private HashMap			interfaces = new HashMap();
    
    void setSenderFactory(PacketSenderFactory f) {
	sender_factory = f;
    }

    static void setDefaultSenderFactory(PacketSenderFactory f) {
	default_factory = f;
    }

    //
    // Note on Concurrency Control:
    //
    // in general, all these methods should be synchronized, but in
    // this case I prefer to control concurrency directly at the
    // HierarchicalDispatcher level.
    //
    public Interface get(String address) throws InvalidSenderException {
	Interface i = (Interface)interfaces.get(address);
	if (i == null) {
	    PacketSender s = sender_factory.createPacketSender(address);
	    i = new Interface(address.getBytes(), s);
	    interfaces.put(address, i);
	}
	return i;
    }

    public Interface get(byte [] address) throws InvalidSenderException {
	String address_s = new String(address);
	Interface i = (Interface)interfaces.get(address_s);
	if (i == null) {
	    PacketSender s = sender_factory.createPacketSender(address_s);
	    i = new Interface(address, s);
	    interfaces.put(address_s, i);
	}
	return i;
    }

    public void cleanup() {
	Map.Entry e;
	for(Iterator i = interfaces.entrySet().iterator(); i.hasNext();) {
	    e = (Map.Entry)i.next();
	    Interface interf = (Interface)e.getValue();
	    if (interf.has_no_refs()) {
		try {
		    interf.shutdown();
		} catch (PacketSenderException ex) {
		    Logging.prlnerr("error closing packet sender "
				    + interf.toString() + ": "
				    + ex.toString());
		}
		Logging.prlnlog("interfaces cleanup: removing interface " + e);
		i.remove(); 
		// assuming that this actually removes the
		// mapping from the original map---not
		// from the set returned by entrySet()
	    }
	}
    }
}

/** implementation of a Siena event notification service.
 *
 *  This is the primary implementation of the Siena event notification
 *  service. A <code>HierarchicalDispatcher</code> can serve as a
 *  Siena event service for local (same Java VM) clients as well as
 *  remote clients.  <code>HierarchicalDispatcher</code>s can also be
 *  combined in a distributed architecture with other dispatchers.
 *  Every dispatcher can be connected to a <em>master</em> dispatcher,
 *  thereby forming a hierarchical structure.  The hierarchy of
 *  dispatchers is assembled incrementally as new dispatchers are
 *  created and connected to a master that already belongs to the
 *  hierarchy.
 *  
 *  <p>A <code>HierarchicalDispatcher</code> uses a {@link
 *  PacketReceiver} to receive notifications, subscriptions and
 *  unsubscriptions from external clients and from its <em>master</em>
 *  dispatcher.  In order to receive and process external requests, a
 *  <code>HierarchicalDispatcher</code> can either use a pool of
 *  internal threads, or it can use users' threads.  See {@link
 *  #DefaultThreadCount}, {@link #setReceiver(PacketReceiver)}, and
 *  {@link #setReceiver(PacketReceiver, int)}
 *  
 *  @see Siena
 *  @see ThinClient
 **/
public class HierarchicalDispatcher implements Siena,  Runnable {
    private int				cleanup_max_r	= 1;
    private int 			cleanup_rounds	= 0;
    private long			cleanup_time	= 0;
    private long			cleanup_max_t	= 10000;
    private String			storeFileName	= null;
    private Poset			subscriptions	= new Poset();
    private ContactsTable		contacts	= new ContactsTable();
    private IFManager			ifmanager	= new IFManager();
    static final byte[]			master_id	= { 0x00 };
    private byte[]			master_address	= null;
    private Interface			master_interface= null;
    private PacketReceiver		listener	= null;
    private byte[]			my_identity	= null;
    private List			matchers	= new LinkedList();

    private SENPPacket			spkt = new SENPPacket();
    private byte []			sndbuf = new byte[SENP.MaxPacketLen];
	
	//
	// mdrosou - <mode> defines the timing policy. In case of the periodic timing policy, <period_length> defines the period length
	//(in number of notifications) and <period_counter> counts the published notifications
	//
	public int mode = 0;	// 0: continuous 	1: sliding-window	2: periodic 3: sliding-window2  4: history
	private static int num_of_periods = 1;	// current period number - initialized at 1
	private static int period_counter = 0;
	private static int events_counter = 0;
	public int window_size = 0;
	public int period_length = 0;
	public boolean diversify = false;
	public int kvalue = 0;
	public double sigma = 0.0;	
	String newline = System.getProperty("line.separator");
	private static long matching_time = (long)0.0;
	//private static long sort_time = (long)0.0;
	private static int sort_counter = 0;
	

    /** configures the subscription storage mechanism.
     *
     *  This server will refresh its database of subscriptions after
     *  processing an incoming request, whenever it has processed more
     *  than <em>n</em> subscription or unsubscription requests, or
     *  <em>t</em> milliseconds have passed since the last refresh.
     *  requests that may be processed without refreshing the
     *  subscription storage.
     * 
     *  <p>The default timeout is 10000 (i.e., 10 seconds).
     **/
    synchronized public void setStoreRefreshTimeout(long t) {
	cleanup_max_t = t;
    }
    
    /** configures the subscription storage mechanism.
     *
     *  Determines the maximum number of subscription or
     *  unsubscription requests that may be processed without
     *  refreshing the subscription store.  See {@link
     *  #setStoreRefreshTimeout(long)} for a more formal description
     *  of the semantics of this parameter.
     * 
     *  <p>The default counter value is 5.
     **/
    synchronized public void setStoreRefreshCounter(int c) {
	cleanup_max_r = c;
    }
    
    /** initializes the subscription storage mechanism.
     *
     *  The server implements a periodic persistent storage mechanism
     *  for subscriptions and contact information for remote clients.
     *  This method defines the storage file name.  If a storage file
     *  already exists, the server attempts to load previously saved
     *  subscriptions from that file.  This mechanism is periodic in
     *  the sense that the server flushes its tables to a file at
     *  configurable intervals.  See {@link
     *  #setStoreRefreshTimeout(long)} and {@link
     *  #setStoreRefreshCounter(int)} for information on how to
     *  configure the refresh rate of the storage mechanism.
     * 
     *  @exception IOException on errors opening the given store file	   
     * 
     *  @exception SienaException wrong format for the store file 
     **/
    synchronized public void initStore(String fn) 
	throws IOException, SienaException {
	storeFileName = fn;
	File f = new File(storeFileName);
	if (!f.createNewFile())
	    loadSubscriptions(storeFileName);
    }

    /** saves remote subscriptions and client information to a file.
     *
     *  Saves the all the contact information regarding remote
     *  subscribers, and all their subscriptions into the given file.
     *  This method is implicitly called at configurable intervals by
     *  the storage function of the server.  See {@link
     *  #initStore(String)} for information on how to initialize the
     *  automatic storage mechanism, and {@link
     *  #setStoreRefreshTimeout(long)} and {@link
     *  #setStoreRefreshCounter(int)} for information on how to
     *  configure the refresh rate of the storage mechanism.
     * 
     *  @exception IOException on errors writing into the given store
     *             file
     **/
    synchronized public void saveSubscriptions(String fname) 
	throws IOException {
	// mdrosou - adding "true" argument to append instead of overwrite
	SENPWriter out = new SENPWriter(new FileOutputStream(fname, true));
	contacts.writeTo(out);
	subscriptions.writeTo(out);
	out.close();
    }

    /** loads remote subscriptions and client information from a file.
     *
     *  This method can be called to load subscriptions and contact
     *  information from a file.  This method is implicitly called by
     *  {@link #initStore(String)} if a storage file already exists.
     * 
     *  @exception IOException on errors reading from the given store
     *             file
     * 
     *  @exception SienaException wrong format for the store file 
     **/
    
	synchronized public void loadSubscriptions(String fname) 
	throws IOException, SienaException {
	BufferedReader in = new BufferedReader(new FileReader(fname));
	SENPBuffer sb = new SENPBuffer();
	String inl = in.readLine();
	int lineno = 1;

	while (inl != null) {
	    switch (inl.charAt(0)) {
	    case 'I': {
		String iname = inl.substring(1);
		inl = in.readLine();
		lineno++;
		if (inl == null || inl.charAt(0) != 'H') 
		    throw new BadStorageFormat(lineno + ": expecting 'H' line in subscription store");
		String handler = inl.substring(1);
		contacts.put(new RemoteSubscriber(iname.getBytes(), 
						  ifmanager.get(handler), kvalue));
		break;
	    }
	    case 'F': {
		sb.init(inl.substring(1).getBytes());
		Filter f;
		try {
		    f = sb.decodeFilter();
		} catch (SENPInvalidFormat ex) {
		    throw new BadStorageFormat(lineno + ": " + ex.getMessage());
		}
		while((inl = in.readLine()) != null && inl.charAt(0) == 'S') {
		    lineno++;
		    //
		    // sanity check here
		    //
		    Subscriber s = contacts.get(inl.substring(1));
		    if (s == null) {
			throw new BadStorageFormat(lineno 
						   + ": unknown subscriber: " 
						   + inl.substring(1));
		    } else {
			
			// mdrosou - I should modify the parser to also read the scores but since
			// I won't use this feature I'll leave it for now
			subscribe(f, s, null);
		    }
		}
		lineno++;

		//
		// here I read either an eof or a non-S line, so I
		// must go back to the main loop
		//
		continue;
	    }
	    default:
		throw new BadStorageFormat(lineno + 
					   ": expecting subscription store");
	    }
	    inl = in.readLine();
	    lineno++;
	}
    }

    /** sets the packet-sender factory associated with this
     *  HierarchicalDispatcher
     *
     *  @see #setDefaultPacketSenderFactory(PacketSenderFactory)
     **/
    synchronized public void setPacketSenderFactory(PacketSenderFactory f) {
	ifmanager.setSenderFactory(f);
    }

    /** default packet-sender factory for HierarchicalDispatcher
     *  interfaces
     *
     *  every new HierarchicalDispatcher objects is assigned this
     *  factory
     *
     *  @see #setPacketSenderFactory(PacketSenderFactory)
     **/
    static public void setDefaultPacketSenderFactory(PacketSenderFactory f) {
	IFManager.setDefaultSenderFactory(f);
    }

    /** default number of threads handling external requests.
     *
     *  Every HierarchicalDispatcher creates a pool of threads to read
     *  and process incoming requests.  This parameter determines the
     *  default number of threads in the pool.  The initial default
     *  value is 5. See {@link #setReceiver(PacketReceiver, int)} for
     *  the semantics of this value.  <code>DefaultThreadCount</code>
     *  is used to create threads upon call to {@link
     *  #setReceiver(PacketReceiver)}.
     *
     *	@see #setReceiver(PacketReceiver) 
     *	@see #setReceiver(PacketReceiver,int) 
     *	@see #setMaster(String)
     **/
    public  int			DefaultThreadCount = 5;

    /** number of failed notifications before a subscriber is
     *	implicitly disconnected.
     *
     *	The default value of <code>MaxFailedConnectionNumber</code> is 2.
     *
     *  HierachicalDispatcher implements a garbage collection
     *  mechanism for unreachable subscribers.  This mechanism
     *  implicitly unsubscribes a client when the dispatcher fails to
     *  connect to the client for a given number of times and after a
     *  given number of milliseconds.  More formally, the dispatcher
     *  considers the sequence of consecutive failed connections not
     *  followed by any successful connection.  This sequence is
     *  charachterized by two parameters: its <em>length</em> and its
     *  <em>duration</em>.<p>
     *
     *	<code>MaxFailedConnectionsNumber</code> represents the upper
     *	bound to the length of the sequence, while
     *	<code>MaxFailedConnectionsDuration</code> represents the upper
     *	bound to the duration of the sequence.  For both parameters, a
     *	negative value means
     *	<em>infinity</em>. <code>removeUnreachableSubscriber()</code>
     *	removes all the subscriptions of those subscribers that have
     *	not been reachable for more than
     *	<code>MaxFailedConnectionsNumber</code> times and more than
     *	<code>MaxFailedConnectionsDuration</code> milliseconds.
     *	Formally, a subscriber that has not been reachable for
     *	<em>T</em> milliseconds for <em>N</em> notifications will be
     *	removed according to the following conditions:
     *
     *  <pre><code>
     *	if (MaxFailedConnectionsNumber &gt;= 0 
     *      &amp;&amp; MaxFailedConnectionsDuration &gt;= 0) {
     *      if (T &gt; MaxFailedConnectionsDuration 
     *          &amp;&amp; N &gt; MaxFailedConnectionsNumber)
     *          remove it!
     *	} else if (MaxFailedConnectionsNumber &gt;= 0) {
     *      if (N &gt; MaxFailedConnectionsNumber)
     *          remove it!
     *	} else if (MaxFailedConnectionsDuration &gt;= 0) {
     *	    if (T &gt; MaxFailedConnectionsDuration)
     *		remove it!
     *	}
     *  </code></pre>
     *
     *	@see #MaxFailedConnectionsDuration
     **/
    public  int			MaxFailedConnectionsNumber	= 2;

    /** milliseconds before automatic unsubscription is activated.
     *
     *  The default value of <code>MaxFailedConnectionsDuration</code>
     *  is 5000 (i.e., 5 seconds).
     *  
     *  @see #MaxFailedConnectionsNumber 
     **/
    public  long		MaxFailedConnectionsDuration	= 5000;
    
    /** creates a dispatcher with a specific <em>identity</em>.  
     *
     *	Every object involved in Siena communications must have a
     *	unique identity.  Strictly speaking, with the hierarchical
     *	protocol (used by this implementation) the identity of an
     *	object must be unique within the group of clients of the same
     *	master server.  However, because client-master connections can
     *	be dynamically reconfigured, it is best to use globally-unique
     *	identities.
     *
     *  <p>Also, notice that the hierarchical protocol reserves the
     *  identity "\000" for its internal use.  Therefore, the string
     *  "\000" must never be used as an identity.
     *
     *	@param id identity given to the dispatcher.  
     **/
    public HierarchicalDispatcher(String id) {
	my_identity = id.getBytes();
	Monitor.add_node(my_identity, Monitor.SienaNode);
    }

    /** creates a dispatcher.
     **/
    public HierarchicalDispatcher() {
	my_identity = SienaId.getId().getBytes();
	Monitor.add_node(my_identity, Monitor.SienaNode);
    }

    /** process a single request, using the caller's thread.
     *
     *  The default value of <code>MaxFailedConnectionsDuration</code>
     *  is 5000 (i.e., 5 seconds).
     *  
     *  @see #DefaultThreadCount
     *  @see #setReceiver(PacketReceiver) 
     *  @see #setReceiver(PacketReceiver, int) 
     **/
    public void processOneRequest() throws SienaException {
	SENPPacket req = SENPPacket.allocate();
	try {
	    int res;
	    req.init();
	    res = listener.receive(req.buf);
	    Logging.prlnlog(new String(req.buf, 0, res));
	    req.init(res);
	    req.decode();
	    if (req != null) processRequest(req);
	} finally {
	    SENPPacket.recycle(req);
	}
    }
    
    public void run() {
	SENPPacket req = SENPPacket.allocate();
	int res;
	while(true) {
	    try {
		res = listener.receive(req.buf);
		req.init(res);
		req.decode();
		processRequest(req);
	    } catch (SENPInvalidFormat ex) {
		Logging.prlnerr("invalid request: " + ex.toString());
	    } catch (PacketReceiverClosed ex) {
		if (ex.getIOException() != null) 
		    Logging.prlnerr("error in packet receiver: " 
				    + ex.toString());
		SENPPacket.recycle(req);
		return;
	    } catch (PacketReceiverFatalError ex) {
		Logging.prlnerr("fatal error in packet receiver: " 
				+ ex.toString());
		SENPPacket.recycle(req);
		return;
	    } catch (PacketReceiverException ex) {
		//
		// non fatal error: just log it and loop
		//
		Logging.prlnlog("non-fatal error in packet receiver: " 
				+ ex.toString());
	    }
	}
    }

    synchronized void removeUnreachableSubscriber(Subscriber sub) {
	if (MaxFailedConnectionsNumber < 0 && MaxFailedConnectionsDuration < 0)
	    return;

	boolean to_remove = false;
	if (MaxFailedConnectionsNumber >= 0 
	    && MaxFailedConnectionsDuration >= 0) {
	    //
	    // pay attention to both time and count (conjunction)
	    //
	    if (sub.getMillisSinceGood() > MaxFailedConnectionsDuration
		&& sub.getFailedAttempts() > MaxFailedConnectionsNumber)
		to_remove = true;
	} else if (MaxFailedConnectionsNumber >= 0) {
	    if (sub.getFailedAttempts() > MaxFailedConnectionsNumber)
		to_remove = true;
	} else if (MaxFailedConnectionsDuration >= 0) {
	    if (sub.getMillisSinceGood() > MaxFailedConnectionsDuration)
		to_remove = true;
	}
	if (to_remove) {
	    Logging.prlnlog("removing unreachable subscriber " 
			    + sub.toString());
	    unsubscribe(null, sub, null);
	}
    }

    /** sets the <em>packet receiver</em> for this server.  
     *
     *	A <em>packet receiver</em> accepts notifications,
     *	subscriptions, and other requests on some communication
     *	channel.  <code>setReceiver</code> will shut down any
     *	previously activated receiver for this dispatcher.  This
     *	method does not guarantee a transactional switch to a new
     *	receiver.  This means that some requests might get lost while
     *	the server has closed the old port and before it reopens the
     *	new port.
     *
     *  <p>This method simply calls {@link #setReceiver(PacketReceiver,
     *  int)} using {@link #DefaultThreadCount} as a default value.
     *
     *	@param r is the receiver 
     *      		   
     *  @see #shutdown()
     *  @see #setReceiver(PacketReceiver, int) 
     **/
    public void setReceiver(PacketReceiver r) {
	setReceiver(r, DefaultThreadCount);
    }

    /** sets the <em>packet receiver</em> for this server.  
     *
     *	A <em>packet receiver</em> accepts notifications,
     *	subscriptions, and other requests on some communication
     *	channel.  <code>setReceiver</code> will shut down any
     *	previously activated receiver for this dispatcher.  This
     *	method does not guarantee a transactional switch to a new
     *	receiver.  This means that some requests might get lost while
     *	the server has closed the old port and before it reopens the
     *	new port.
     *       
     *	@param r the packet receiver
     *  @param threads is the number of threads associated with the
     *  receiver, and therefore to the whole server.  A positive value
     *  causes this dispatcher to create threads.  A value of 0 causes
     *  the dispatcher not to create any thread, In this case, the
     *  application must explicitly call {@link #processOneRequest()}.
     *
     *  @see #shutdown()
     *  @see #setMaster(String) 
     **/
    synchronized public void setReceiver(PacketReceiver r, int threads) {
	if (listener != null) {
	    try {
		listener.shutdown();
	    } catch (PacketReceiverException ex) {
		Logging.exerr(ex);
		//
		// ...work in progress...
		//
	    }
	    //
	    // this should send a PacketReceiverClosed exception to
	    // every thread that is waiting for packets on the old
	    // listener, which will make them exit normally.  However,
	    // because of bugs in the JVM, or because of bad
	    // implementations of packetReceiver, this might not be
	    // true.  ...work in progress...
	    //
	}
	listener = r;

	if (master_interface != null) {
	    spkt.init();
	    spkt.method = SENP.MAP;
	    spkt.id = my_identity;
	    spkt.handler = listener.address();
	    try {
		master_interface.send(spkt.buf, spkt.encode());
	    } catch (Exception ex) {
		Logging.prlnerr("error sending MAP packet to master " 
				+ master_interface.toString() 
				+ ": " + ex.toString());
		Logging.exerr(ex);
		//
		// I should really do something here
		// ...work in progress...
		//
	    }
	}
	//
	// now fires off the reader threads for this dispatcher
	//
	while (threads-- > 0) {
	    Thread t = new Thread(this);
	    //
	    // Perhaps I should set t as a deamon thread...
	    //
	    t.start();
	}
    }
    
    /** creates a TCP-based <em>packet receiver</em> for this server.
     *
     *	@deprecated as of Siena 1.1.0, replaced by {@link
     *              #setReceiver(PacketReceiver) setReceiver()}
     *
     *  @param port is the port number allocated by the listener,
     *	            <code>0</code> allocates a random available port
     *	            number, the default value for Siena is
     *	            <code>SENP.DEFAULT_PORT</code>
     *      		   
     *  @exception IOException when the given port is already in use	   
     **/
    synchronized public void setListener(int port) throws IOException {
	setReceiver(new TCPPacketReceiver(port));
    }

    /** connects this dispatcher to a <em>master</em> dispatcher.
     *	
     *	If this dispatcher is already connected to a master
     *	dispatcher, <code>setMaster</code> disconnects the old one and
     *	connects the new one, thereby unsubscribing all the top-level
     *	subscriptions and resubscribing with the new one.  This method
     *	should be used only when this dispatcher <em>needs</em> to
     *	switch to a different master, it is not necessary (it is in
     *	fact very inefficient) to set the master before every
     *	subscription or notification.
     *
     *	<p>This method does not guarantee a transactional switch.  This
     *	means that some notifications might be lost when the server
     *	has detached from the old master and before it re-subscribes
     *	with the new master_interface.
     *
     *	<p>If this dispatcher does not have a <em>packet receiver</em>
     *	associated with it, <code>setMaster</code> implicitly sets up
     *	one for the dispatcher.  The default receiver is a
     *	<code>TCPPacketReceiver</code> listening to a randomly
     *	allocated port.  If you are not happy with the default
     *	decision, you should call <code>setReceiver()</code> before
     *	you call <code>setMaster</code>.
     *  
     *  <p>When <code>address</code> is null, simply disconnect from the
     *  current master server.
     *
     *	@param address is the external identifier of the master dispatcher
     *	           (e.g., <code>"ka:host.domain.edu:8765"</code>),
     *             or null to disconnect from the ccurrent master server.
     *  @see #setReceiver(PacketReceiver)
     *  @see #shutdown() 
     **/
    synchronized public void setMaster(String address) 
	throws InvalidSenderException, java.io.IOException, 
	       PacketSenderException {

	if (address == null) {
	    disconnectMaster();
	    return;
	}

	byte [] new_address = address.getBytes();
	Interface new_interface = ifmanager.get(new_address);

	disconnectMaster();
	boolean new_listener = false;
	if (listener == null) {
	    setReceiver(new TCPPacketReceiver(0));
	    new_listener = true;
	}
	if (master_interface != null) 
	    master_interface.remove_ref();
	master_interface = new_interface;
	master_interface.add_ref();
	master_address = new_address;
	//
	// sends all the top-level subscriptions to the new master
	//
	for(Iterator i = subscriptions.rootsIterator(); i.hasNext();) {
	    Subscription s = (Subscription)i.next();
	    try {
		spkt.init();
		spkt.method = SENP.SUB;
		//
		// mdrosou - here I should choose a score to assosiate with the subscription but
		// since non-Access Points receive all notifications it probably doesn't matter
		//
		spkt.score = -500.0;
		spkt.ttl = SENP.DefaultTtl;
		spkt.id = my_identity;
		spkt.handler = listener.address();
		spkt.filter = s.filter;
		master_interface.send(spkt.buf, spkt.encode());
	    } catch (Exception ex) {
		Logging.prlnerr("error sending SUB packet to master "
				+ master_interface.toString()
				+ ": " + ex.toString());
		Logging.exerr(ex);
		//
		// of course I should do something here...
		// ...work in progress...
		//
	    }
	}
	
	//
	// mdrosou - define <k> value (send it to the master)
	// k = 0 means that this server has other subscribers (is an indemediate server) and so
	// its master should foerward all notifications to it
	//
	try {
		// mdrosou - here I could send this message to the Monitor (I should also modify that program though)
		spkt.init();
		spkt.method = SENP.KDEF;
		spkt.id = my_identity;
		spkt.handler = listener.address();
		spkt.k = 0;
		master_interface.send(spkt.buf, spkt.encode());
	    } catch (Exception ex) {
		Logging.prlnerr("error sending KDEF packet to master "
				+ master_interface.toString()
				+ ": " + ex.toString());
		Logging.exerr(ex);
		//
		// of course I should do something here...
		// ...work in progress...
		//
	    }
	
    }

    /** suspends the connection with the <em>master</em> server of
     *	this dispatcher.  
     *
     *	This causes the <em>master</em> server to stop sending
     *	notification to this dispatcher.  The master correctly
     *	maintains all the existing subscriptions so that the flow of
     *	notification can be later resumed (see {@link #resumeMaster()
     *	resumeMaster}.  This operation can be used when this
     *	dispatcher, that is this virtual machine, is going to be
     *	temporarily disconnected from the network or somehow
     *	unreachable from its master server.
     *
     *	@see #resumeMaster() 
     **/
    synchronized public void suspendMaster() {
	try {
	    spkt.init();
	    spkt.method = SENP.SUS;
	    spkt.to = master_address;
	    spkt.id = my_identity;
	    spkt.handler = listener.address();
	    master_interface.send(spkt.buf, spkt.encode());
	} catch (Exception ex) {
	    Logging.prlnerr("error sending SUS packet to master " 
			    + master_interface.toString()
			    + ": " + ex.toString());
	    //
	    // of course I should do something here...
	    // ...work in progress...
	    //
	}
    }

    /** resumes the connection with the <em>master</em> server.
     *
     *	This causes the <em>master</em> server to resume sending
     *	notification to this dispatcher.
     *
     *	@see #suspendMaster()
     **/
    synchronized public void resumeMaster() {
	try {
	    spkt.init();
	    spkt.method = SENP.RES;
	    spkt.to = master_address;
	    spkt.id = my_identity;
	    spkt.handler = listener.address();
	    master_interface.send(spkt.buf, spkt.encode());
	} catch (Exception ex) {
	    Logging.prlnerr("error sending RES packet to master " 
			    + master_interface.toString()
			    + ": " + ex.toString());
	    //
	    // of course I should do something here...
	    // ...work in progress...
	    //
	}
    }

    /** returns the identity of this dispatcher.
     *
     *  every object in a Siena network has a unique identifier.  This
     *  method returns the identifier of this dispatcher.  
     *
     *  @see #HierarchicalDispatcher(String)
     **/
    synchronized public String getIdentity() {
	return new String(my_identity);
    }

    /** returns the address of the master server associated
     *	with this dispatcher.
     *     
     *  @return address of the master server or <code>null</code> if the
     *		master server is not set
     *
     *  @see #setMaster(String)
     **/
    synchronized public String getMaster() {
	if (master_address == null) return null;
	return new String(master_address);
    }

    /** returns the listener associated with this dispatcher.
     *
     *	@return receiver of this dispatcher (possibly null)
     *     
     *  @see #setReceiver(PacketReceiver)
     **/
    synchronized public PacketReceiver getReceiver() {
	return listener;
    }

    synchronized private void disconnectMaster() {
	if (master_interface != null) {
	    try {
		spkt.init();
		spkt.method = SENP.BYE;
		spkt.id = my_identity;
		spkt.to = master_address;
		master_interface.send(spkt.buf, spkt.encode());
	    } catch (PacketSenderException ex) {
		Logging.prlnerr("error sending BYE packet to master " 
				+ master_interface.toString() 
				+ ": " + ex.toString());
		//
		// well, what would you do in this case?
		// ...work in progress...
		//
	    }
	    master_interface = null;
	    master_address = null;
	}
    }

    private void processRequest(SENPPacket req) {
	// Logging.prlnlog("processRequest: " + req);
	if (req == null) {
	    Logging.prlnerr("processRequest: null request");
	    return;
	}
		
	if (req.ttl <= 0) return;
	req.ttl--;
	try {
	    switch(req.method) {
	    case SENP.PUB: publish(req); break;
	    case SENP.SUB: subscribe(req); ++cleanup_rounds; break;
	    case SENP.BYE: req.pattern = null; req.filter = null;
	    case SENP.UNS: unsubscribe(req); ++cleanup_rounds; break;
	    case SENP.SUS: suspend(req); ++cleanup_rounds; break;
	    case SENP.RES: resume(req); ++cleanup_rounds; break;		
	    case SENP.MAP: map(req); ++cleanup_rounds; break;
	    case SENP.CNF: configure(req); break;
		
		// mdrosou - adding new types of packets
		case SENP.STATS: collectStats(req); break;
		case SENP.KDEF: setK(req); break;
		case SENP.WRITE: writeToFile(new String(req.write)); break;
		
	    case SENP.OFF: 
		shutdown(); 
		//
		// BEGIN_UNOFFICIAL_PATCH
		try { Thread.sleep(500); } catch (Exception ex) {};
		System.exit(0);
		// END_UNOFFICIAL_PATCH
		//
		break;
	    case SENP.NOP: break;
	    default:
		Logging.prlnerr("processRequest: unknown method: " + req);
		//
		// can't handle this request (yet)
		// ...work in progress...
		//
	    }
	    //
	    // here I see if it's time to clean up tables and save
	    // subscriptions
	    //
	    long current_time = System.currentTimeMillis();
	    if (cleanup_rounds >= cleanup_max_r ||
		current_time - cleanup_time > cleanup_max_t) {
		//Logging.prlnlog("cleaning up contacts and interfaces tables");
		synchronized (this) {
		    ifmanager.cleanup();
		    contacts.cleanup();
		    if (storeFileName != null)
			saveSubscriptions(storeFileName);
		    cleanup_time = current_time;
		    cleanup_rounds = 0;
		}
	    }
	} catch (Exception ex) {
	    Logging.exerr(ex);
	    //
	    // log something here ...work in progress...
	    //
	}
    }
    
    synchronized private void configure(SENPPacket req) {
	if (req.ttl == 0) return;
	if (req.handler == null) {
	    Logging.prlnlog("reconfigure: disconnecting form master");
	    disconnectMaster();
	} else {
	    try {
		String new_master = new String(req.handler);
		Logging.prlnlog("reconfigure: switching to master " 
				+ new_master);
		setMaster(new_master);
	    } catch (Exception ex) {
		Logging.prlnerr("configure: failed reconfiguration request: " 
				+ ex.toString());
	    }
	}
    }

    private void map(SENPPacket req) {
	if (req.id == null || req.ttl == 0 || req.handler == null) return;
	synchronized (this) {
	    Subscriber s = contacts.get(new String(req.id));
	    if (s != null) 
		try {
		    if (!s.isLocal())
			s.mapHandler(ifmanager.get(req.handler));
		} catch (InvalidSenderException ex) {
		    Logging.prlnerr("error while mapping handler " 
				    + new String(req.handler) 
				    + ": " + ex.toString());
		    Logging.prlnerr("client not remapped");
		}
	}
    }

    private void suspend(SENPPacket req) {
	if (req.id == null || req.ttl == 0) return;
	synchronized(this) {
	    //
	    // I'm not 100% sure this synchronization is really necessary.
	    //
	    Subscriber s = contacts.get(new String(req.id));
	    if (s != null) s.suspend();
	}
    }

    private void resume(SENPPacket req) {
	if (req.id == null || req.ttl == 0) return;
	synchronized(this) {
	    //
	    // I'm not 100% sure this synchronization is really necessary.
	    //
	    Subscriber s = contacts.get(new String(req.id));
	    if (s != null) s.resume();
	}
    }

    private void publish(SENPPacket req) {
	if (req.event == null) {
	    Logging.prlnerr("Warning: null event in PUB message: "+req);
	    return;
	}	
	byte[] sender = req.id;
	if (sender != null && !SENP.match(sender, master_id)) {
	    Monitor.notify(sender, my_identity);
	}
	//
	// first forward to master
	//
	if (master_interface != null  && req.ttl > 0 && 
	    (sender == null || !SENP.match(sender, master_id))) {
	    req.id = my_identity;
	    req.to = master_id;
	    try {
		master_interface.send(req.buf, req.encode());
	    } catch (Exception ex) {
		Logging.prlnerr("error sending PUB packet to master " 
				+ master_interface.toString()
				+ ": " + ex.toString());
	    }
	}
	req.id = master_id;
	
	//
	// mdrosou - the rest of this fuction differs a lot from the original code - it also works only for remote subscribers (not local ones)
	//
	period_counter++;
	events_counter++;
	
	// mdrosou - find matching subscribers 
	SSet matching;
	synchronized (this) {		
		long startTimer = System.nanoTime();	
		
		matching = subscriptions.matchingSubscribers(req.event);		
		
		//Logging.prlnerr(new Integer(subscriptions.num_checked_nodes).toString());
		
		long endTimer = System.nanoTime();
		matching_time += endTimer - startTimer;
		// long curr_avg = (long)(matching_time/events_counter);
		// Logging.prlnerr(newline + "Matching time for " + req.event + ": " + new Long(endTimer - startTimer).toString() + ", current avg: " + new Long(curr_avg).toString());
	}
	
	SubscriberIterator i = matching.iterator();
	
	// mdrosou - continuous and sliding-window modes	
	if(mode == 0 || mode == 1 || mode == 3 || mode == 4) {				
		// mdrosou - find all subscribers, matching or not
		Iterator ii;
		synchronized (this) {
			ii = contacts.contacts.entrySet().iterator();
		}

		// create dummy notification (unique for each event)
		Notification e = new Notification();
		e.putAttribute("Dummy", events_counter);	
	
		// mdrosou - check all subscribers to see if the new event should be sent to them
		while(ii.hasNext()) {
		    Subscriber s =  (Subscriber)((Map.Entry)ii.next()).getValue();
		    byte [] s_identity = s.getIdentity();
		    
			// if the subscriber was not a matching one, add a dummy notification to its buffer
			// set <already_sent> field to true so that it'll never be forwarded to the subscriber		
			if(!matching.contains(s)) {
				CandidateNotif ee = new CandidateNotif(-1.0, e);
				ee.already_sent = true;
				s.CandidateNotifications.addFirst(ee);
			}			
			
			if (s.isLocal()) {
			//
			// always notifies local subscribers
			//
			Monitor.notify(my_identity, s_identity);
			if (!s.notify(req))
			    removeUnreachableSubscriber(s);
		    } else if (req.ttl > 0 && (sender == null || !SENP.match(sender, s_identity))) {			
				// mdrosou - this is the remote subscriber case
				// in continuous  mode <n> has 0 or 1 results. In sliding-window mode it can have more when using diversity	
				CandidateNotif[] n;
				synchronized (s.CandidateNotifications) {
					// delete oldest event that has left the window
					CandidateNotif expired = null;
					if(events_counter > window_size && s.CandidateNotifications.size() > 0 ) {
						//Logging.prlnerr("\n\nDeleting: " + s.CandidateNotifications.getLast().toString());
						expired = (CandidateNotif)s.CandidateNotifications.getLast();
						s.CandidateNotifications.removeLast();
						if(s.current_results.remove(expired))
							;//Logging.prlnerr("\nDeleted: " + expired.toString());
					}
					// get top-k. It should be either null or containing one result (the newest event)
					//long startsTimer = System.nanoTime();
					if(mode == 0)
						n = s.sortCandidates_continuous(diversify, sigma);
					else if(mode == 1)
						n = s.sortCandidates_slidingWindow(diversify, sigma);
					else if(mode == 2)
						n = s.sortCandidates_slidingWindow(diversify, sigma);
					else if(mode == 3)
						n = s.sortCandidates_slidingWindow2(diversify, sigma);
					else
						n = s.sortCandidates_history(expired, window_size, diversify, sigma);
					//long endsTimer = System.nanoTime();
					//sort_time += endsTimer - startsTimer;
					//sort_counter++;
				}
				
				//Logging.prlnerr(new Integer(n.length).toString());			
				
				for(int j=0; j<n.length; j++) {
					SENPPacket spkt = SENPPacket.allocate();
					if(n[j] != null) {
						spkt.event = n[j].notif;
						spkt.score = n[j].score;
					}
					spkt.method = SENP.PUB; 
					spkt.id = my_identity;
					
					//Logging.prlnerr("Sending: " + spkt.toString());
				
					spkt.to = s_identity;
					if (s_identity != null)
						Monitor.notify(my_identity, s_identity);
				
					if (spkt.event == null) {
						;////Logging.prlnerr("Warning: null event in PUB message: "+spkt);
						return;
					}	
				
					try { 
						if (!s.notify(spkt))
							removeUnreachableSubscriber(s);
					} finally {
						SENPPacket.recycle(spkt);
					}		
			    }
			}
		}
	}	
	// mdrosou - periodic mode
	// if period_counter < period_length do nothing. Events have are added to matching buffers when we called the matching subscribers function
	else if(mode == 2 && period_counter >= period_length) {				
		String newline = System.getProperty("line.separator");
		Logging.prlnerr("Period number " + num_of_periods + " (notif " + period_counter*num_of_periods + ") " + diversify + " " + sigma);
		num_of_periods++;
		period_counter = 0;	
	
		// mdrosou - get all subscribers, matching or not
		Iterator ii;
		synchronized (this) {
			ii = contacts.contacts.entrySet().iterator();
		}	
	
		// mdrosou - ask all subscribers for their top-k results
		while(ii.hasNext()) {
		    Subscriber s =  (Subscriber)((Map.Entry)ii.next()).getValue();
		    byte [] s_identity = s.getIdentity();
		    
			if (s.isLocal()) {
			//
			// always notifies local subscribers
			//
			Monitor.notify(my_identity, s_identity);
			if (!s.notify(req))
			    removeUnreachableSubscriber(s);
		    } else if (req.ttl > 0 && (sender == null || !SENP.match(sender, s_identity))) {			
				// mdrosou - this is the remote subscriber case
				// all subscribers are asked for their top-k results		
				CandidateNotif[] n;
				synchronized (this) {
					//long startsTimer = System.nanoTime();
					n = s.sortCandidates_periodic(diversify, sigma);
					//long endsTimer = System.nanoTime();
					//sort_time += endsTimer - startsTimer;
					//sort_counter++;
				}
				
				//Logging.prlnerr(new Integer(n.length).toString());				
				
				for(int j=0; j<n.length; j++) {
					SENPPacket spkt = SENPPacket.allocate();
					if(n[j] != null) {
						spkt.event = n[j].notif;
						spkt.score = n[j].score;
					}
					spkt.method = SENP.PUB; 
					spkt.id = my_identity;
					
					//Logging.prlnerr("Sending: " + spkt.toString());
				
					spkt.to = s_identity;
					if (s_identity != null)
						Monitor.notify(my_identity, s_identity);
				
					if (spkt.event == null) {
						//Logging.prlnerr("Warning: null event in PUB message: "+spkt);
						return;
					}	
		
					try { 
						if (!s.notify(spkt))
							removeUnreachableSubscriber(s);
					} finally {
						SENPPacket.recycle(spkt);
					}		
			    }
			}
		}
	}	
    }
	
	//
	// mdrosou - collect statistics about the server's poset
	// Since it is synchronized all statistics are gathered independently 
	//
	synchronized private void collectStats(SENPPacket req) 
	throws InvalidSenderException, SienaException {
	
	SENPPacket spkt = SENPPacket.allocate();
	spkt.method = SENP.STATS; 
	spkt.id = my_identity;
	spkt.handler = listener.address();
	spkt.server_index = req.server_index;
	spkt.to = req.id;
	
	// print subsbscriptions in the STATS message
	Logging.prlnlog("STATS has these subscriptions: ");	
	
	for(int i=0;i<req.pattern.filters.length;i++)
		Logging.prlnlog(req.pattern.filters[i].toString());	
		
	// how many new roots will be created
	spkt.num_of_new_roots = subscriptions.countNewRoots(req.pattern);
	
	// how many ncovered odes in the poset
	spkt.num_of_covered_nodes = subscriptions.countCoveredSubscriptions(req.pattern);
	
	// how many nodes in the poset
	spkt.num_of_poset_nodes = subscriptions.countSubscriptions();	
	
	Logging.prlnlog("sending STATS: " + spkt);	
	
	// 2nd way of notifying
	// Subscriber s = map_subscriber(req);
	// if (s == null) {
	// Logging.prlnerr("subscribe: unknown subscriber: " 
			// + req.toString());
	// return;
	// }
	
	//if (!s.notify(spkt))
	   // removeUnreachableSubscriber(s);
	
	// send statistics to the client that asked
	Interface new_interface = ifmanager.get(req.handler);
	
	try {
		// here I could send this message to the Monitor (I should also modify that program though)
		new_interface.send(spkt.buf, spkt.encode());
	} catch (Exception ex) {
		Logging.prlnerr("error sending STATS packet to interface "
				+ new_interface.toString()
				+ ": " + ex.toString());
		Logging.exerr(ex);
    }

	return;
	}
	
	// mdrosou - set <k> for the subscriber who sent <req>
	private void setK(SENPPacket req)
	throws InvalidSenderException, SienaException {
	
	Subscriber s = map_subscriber(req);
	if (s == null) {
	Logging.prlnerr("subscribe: unknown subscriber: " 
			+ req.toString());
	return;
	}
	
	s.k = req.k;	
	// mdrosou - so that the subscriber is not removed if they are late to send actual subscriptions. I should probably modify addRef to not increase when ref==1 (only if *all* clients send KDEF packets)
	s.addRef();
	
	return;	
	}
	
	// mdrosou - write to file
	public void writeToFile(String filename){
	try{
		FileWriter fstream = new FileWriter(filename);
		BufferedWriter hd_log = new BufferedWriter(fstream);
		
		if(filename.indexOf("matching") != -1) {		
			hd_log.write("Poset size: " + subscriptions.countSize() + newline);
			hd_log.write("Checked nodes: " + subscriptions.num_checked_nodes + newline);
			hd_log.write("Total matching time: " + new Long(matching_time).toString() + newline);
			hd_log.write("Matching time avg: " + new Long(matching_time/(long)events_counter).toString() + newline);
		}
		else {		
			Iterator ii;
			synchronized (this) {
				ii = contacts.contacts.entrySet().iterator();
			}

			while(ii.hasNext()) {
				Subscriber s =  (Subscriber)((Map.Entry)ii.next()).getValue();		
				// hd_log.write("k: " + kvalue + newline + "window: " + window_size + newline + "period: " + period_length + newline + "Subscriber: " + s.getIdentity().toString() + newline + "Diversity avg time: " + new Long(s.total_div_time/s.divcounter).toString() + newline + newline);
				hd_log.write("Subscriber: " + s.getIdentity().toString() + newline);
				Iterator li = s.results_per_window.iterator();
				while(li.hasNext()) {
					Iterator si = ((Set)li.next()).iterator();
					while(si.hasNext()) {
						hd_log.write(((CandidateNotif)si.next()).toString() + newline);
					}
					hd_log.write(newline);
				}				
			}

			// hd_log.write(newline + "Avg Total sorting time: " + sort_time / (double) sort_counter + newline);
			// hd_log.write(newline + "Total sorting time: " + sort_time + newline);
			// hd_log.write(newline + "Sort counter: " + sort_counter + newline);
		}
		hd_log.flush();
		hd_log.close();
    }catch (Exception e){
      System.err.println("Error: " + e.getMessage());
    }
    }


    private void subscribe(SENPPacket req) 
	throws InvalidSenderException, SienaException {
	if (req.filter == null && req.pattern == null) {
	    //
	    // null filters/patterns are not allowed in subscriptions
	    // this is a design choice, we could accept null filters
	    // with the semantics of the universal filter: one that
	    // matches every notification
	    //
	    Logging.prlnerr("subscribe: null filter/pattern in subscription");
	    return;
	}

	Monitor.subscribe(req.id, my_identity);
	synchronized (this) {
	    Subscriber s = map_subscriber(req);
	    if (s == null) {
		Logging.prlnerr("subscribe: unknown subscriber: " 
				+ req.toString());
		return;
	    }
	    if (req.filter != null) {

		subscribe(req.filter, s, req);
	    } else {
		//
		// must be req.pattern != null
		//
		subscribe(req.pattern, s);
	    }
	}
    }

    synchronized private void unsubscribe(Filter f, Subscriber s, 
					  SENPPacket req) {
	//
	// for global unsubscriptions (f == null)
	// removes all the patterns as well
	// 
	if (f == null) unsubscribe((Pattern)null, s);
	Set to_remove = subscriptions.to_remove(f, s);
	if (to_remove.isEmpty()) return;

	if (req == null) req = spkt;

	if (master_interface != null && req.ttl > 0) {
	    req.id = my_identity;
	    req.handler = listener.address();
	    req.to = master_id;
	}

	Set new_roots = null;
	for(Iterator i = to_remove.iterator(); i.hasNext();) {
	    Subscription sub = (Subscription)i.next();
	    if (new_roots == null) new_roots = new HashSet();

	    new_roots.addAll(subscriptions.remove(sub));
	    if (master_interface != null && req.ttl > 0) {
		try {
		    req.method = SENP.UNS;
		    req.filter = sub.filter;
		    master_interface.send(req.buf, req.encode());
		} catch (Exception ex) {
		    Logging.prlnerr("error sending UNS packet to master "
				    + master_interface.toString()
				    + ": " + ex.toString());
		}
	    }
	}

	if (new_roots != null) {
	    if (master_interface != null && req.ttl > 0) {
		for(Iterator ri = new_roots.iterator(); ri.hasNext();) {
		    try {
			req.method = SENP.SUB;
			// mdrosou - dummy score
			req.score = -500.0;
			req.filter = ((Subscription)ri.next()).filter;
			master_interface.send(req.buf, req.encode());
		    } catch (Exception ex) {
			Logging.prlnerr("error sending SUB packet to master "
					+ master_interface.toString() 
					+ ": " + ex.toString());
		    }
		}
	    }
	}
	//
	// here I could avoid removing this contact, since this would
	// be caught sooner or later by the periodic cleanup
	//
	if (s.hasNoRefs()) contacts.remove(s);
    }

    private void unsubscribe(SENPPacket req) 
	throws InvalidSenderException, SienaException {
	if (req.id == null) return;
	synchronized (this) {
	    Subscriber s = contacts.get(new String(req.id));
	    if (s == null) return;
	    if (req.pattern != null) {
		Monitor.unsubscribe(req.id, my_identity);
		unsubscribe(req.pattern, s);
	    } else {
		Monitor.unsubscribe(req.id, my_identity);
		unsubscribe(req.filter, s, req);
	    }
	}
    }

    private Subscriber map_subscriber(SENPPacket req) 
	throws InvalidSenderException {
	if (req.id == null) return null;
	synchronized (this) {
	    Subscriber s = contacts.get(new String(req.id));
	    if (s == null) {
		if (req.handler != null) {
		    s = new RemoteSubscriber(req.id, 
					     ifmanager.get(req.handler), kvalue);
		    contacts.put(s);
		}
	    } else if (req.handler != null) {
		s.mapHandler(ifmanager.get(req.handler));
	    }
	    return s;
	}
    }

    synchronized private Subscriber map_subscriber(Notifiable notifiable) {
	Subscriber s = contacts.get(notifiable);
	if (s == null) {
	    s = new LocalSubscriber(notifiable, kvalue);
	    contacts.put(s);
	}	
	
	return s;
    }

    /** closes this dispatcher.  
     *
     *	If this dispatcher has an active listener then closes the
     *	active listener.  If this dispatcher has a master server, then
     *	cancels every subscription with the master server.
     *	<p>
     *
     *	Some implementations of the Java VM do not respond correctly
     *	to shutdown().  In particular, Sun's jvm1.3rc1-linux is known
     *	not to work correctly.  Sun's jvm1.3-solaris, jvm1.3-win32,
     *	and jvm1.2.2-linux work correctly.
     *
     *	@see #setReceiver(PacketReceiver)
     *	@see #setMaster(String) 
     **/
    synchronized public void shutdown() {
	disconnectMaster();
	if (listener != null)
	    try { 
		listener.shutdown(); 
	    } 
	    catch (PacketReceiverException ex) {
		Logging.prlnerr("error shutting down packet receiver: " 
				+ ex.toString());
	    }
	Monitor.remove_node(my_identity);
	}

    /** removes all subscriptions from any notifiable.  
     *
     *  clears all the subscriptions from local or remote clients.  In
     *  case this server has a master server, it instructs the master
     *  server to clear all the subscriptions associated with this
     *  dispatcher.  This method is also useful in case this
     *  dispatcher re-uses both a listener and an identity from a
     *  previous dispatcher that crashed or did not otherwise shutdown
     *  properly.
     *
     *	@see #shutdown()
     **/
    synchronized public void clearSubscriptions() throws SienaException {
	subscriptions.clear();
	contacts.clear();
	if (master_interface != null) {
	    SENPPacket spkt = SENPPacket.allocate();
	    spkt.method = SENP.BYE;
	    spkt.id = my_identity;
	    spkt.to = master_address;
	    try {
		master_interface.send(spkt.buf, spkt.encode());
	    } finally {
		SENPPacket.recycle(spkt);
	    }
	}
    }

    public void publish(Notification e) throws SienaException {
	SENPPacket spkt = SENPPacket.allocate();
	spkt.event = e;
	spkt.method = SENP.PUB; 
	spkt.id = null;
	try {
	    publish(spkt);
	} finally {
	    SENPPacket.recycle(spkt);
	}
    }

    public void suspend(Notifiable n) throws SienaException {
	if (n == null) return;
	synchronized (this) {
	    Subscriber s = contacts.get(n);
	    if (s != null) s.suspend();
	}
    }

    public void resume(Notifiable n) throws SienaException {
	if (n == null) return;
	synchronized (this) {
	    Subscriber s = contacts.get(n);
	    if (s != null) s.resume();
	}
    }

    public void subscribe(Filter f, Notifiable n) throws SienaException {
	if (f == null) {
	    //
	    // null filters are not allowed in subscriptions this is a
	    // design choice, we could accept null filters with the
	    // semantics of the universal filter: one that matches
	    // every notification
	    //
	    Logging.prlnerr("subscribe: null filter in subscription");
	    return;
	}

	synchronized (this) {
	    Subscriber s = map_subscriber(n);
	    if (s == null) {
		Logging.prlnerr("subscribe: unknown local subscriber");
		return;
	    }
	    subscribe(f, s, null);
	}
    }

    synchronized private void subscribe(Filter f, Subscriber s,  
					SENPPacket req) throws SienaException {
	Subscription sub = subscriptions.insert_subscription(f, s);
	if (sub == null) return;
	
	if (subscriptions.is_root(sub) && master_interface != null
	    && (req == null || req.ttl > 0)) {
	    try { 
		if (req == null) { req = spkt; req.init(); }
		req.method = SENP.SUB;
		req.id = my_identity;
		req.handler = listener.address();
		req.to = master_id;
		req.filter = f;
		
		master_interface.send(req.buf, req.encode());
	    } catch (Exception ex) {
		Logging.prlnerr("error sending SUB packet to master " 
				+ master_interface.toString() 
				+ ": " + ex.toString());
		//
		// log something here ...work in progress...
		//
	    }
	}
    }

	synchronized public void subscribe(Pattern p, Notifiable n)
	throws SienaException {
	Subscriber s = map_subscriber(n);
	if (s == null) return;
	subscribe(p, s);
    }

	synchronized private void subscribe(Pattern p, Subscriber s)
	throws SienaException {
	//
	// warning: this is a naive implementation.  
	// ...work in progress...
	//
	PatternMatcher m;
	
	ListIterator li = matchers.listIterator(); 
	while(li.hasNext()) {
	    m = (PatternMatcher)li.next();
	    if (Covering.covers(m.pattern, p)) {
		if (m.subscribers.contains(s)) 
		    return;
		if (Covering.covers(p, m.pattern)) {
			// mdrosou - dummy score
			ExtSubscriber exts = new ExtSubscriber(s, -500.0);
		    m.subscribers.add(exts);
		    return;
		}
	    } else if (Covering.covers(p, m.pattern)) {
		if (m.subscribers.remove(s)) { 
		    if (m.subscribers.isEmpty()) {
			for(int j = 0; j < m.pattern.filters.length; ++j)
			    unsubscribe(m.pattern.filters[j], m);
			//
			// FIXME: this is a bug!  in fact this
			// li.remove() is executed right before a
			// li.hasNext()---the problem is the screwed
			// up semantics of iterators in Java... which
			// may make sense for a generic iterator, but
			// is incredibly stupid for a list iterator:
			// it should be possible to call li.remove()
			// and assume that li shifted to the next
			// element (so that li.hasNext() would work as
			// expected).  This demonstrates once again
			// that defining collections and iterators
			// with generic classes is not always a good
			// idea ... well, enough flames for now.
			// Anyway, this needs to be fixed!
			//
			li.remove(); 
		    }
		}
	    }
	}

	m = new PatternMatcher(p);
	// mdrosou
	ExtSubscriber exts = new ExtSubscriber(s, -500.0);
	m.subscribers.add(exts);
	for(int i = 0; i < p.filters.length; ++i)
	    subscribe(p.filters[i], m);
	matchers.add(m);
    }

    synchronized public void unsubscribe(Filter f, Notifiable n) {
	Subscriber s = contacts.get(n);
	if (s == null) return;
	unsubscribe(f, s, null);
    } 

    synchronized public void unsubscribe(Pattern p, Notifiable n) {
	Subscriber s = map_subscriber(n);
	if (s == null) return;
	unsubscribe(p,s);
    }

    synchronized private void unsubscribe(Pattern p, Subscriber s) {
	//
	// warning: this is a naive implementation.  
	// ...work in progress...
	//
	PatternMatcher m;
	
	ListIterator li = matchers.listIterator(); 
	while(li.hasNext()) {
	    m = (PatternMatcher)li.next();
	    if (p == null || Covering.covers(p, m.pattern)) {
		if (m.subscribers.remove(s)) {
		    if (m.subscribers.isEmpty()) {
			for(int j = 0; j < m.pattern.filters.length; ++j)
			    unsubscribe(m.pattern.filters[j], m);
			//
			// FIXME: this is a bug!  see same comment above
			//
			li.remove(); 
		    }
		}
	    }
	}
	if (s.hasNoRefs()) contacts.remove(s);
    }

    /** this method has no effect.  
     *
     *	Method not implemented in this Siena server. 
     **/
    public void advertise(Filter f, String id) throws SienaException {};

    /** this method has no effect.  
     *
     *  Method not implemented in this Siena server. 
     **/
    public void unadvertise(Filter f, String id) throws SienaException {};

    /** this method has no effect.  
     *
     *	Method not implemented in this Siena server. 
     **/
    public void unadvertise(String id) throws SienaException {};

    synchronized public void unsubscribe(Notifiable n) throws SienaException {
	unsubscribe((Filter)null, n);
    }
}
